""" :mod: GFAL2_StorageBase

.. module: python

:synopsis: GFAL2 class from StorageElement using gfal2. Other modules can inherit from this use the gfal2 methods.


:TODO: When we are totally migrated to python 3, we can explicitely remove the str cast
  of all the ctx calls. They were added because we would receive unicode from DIRAC https servers
  and boost python does not play well with that.

Environment Variables:

* DIRAC_GFAL_GRIDFTP_ENABLE_IPV6: this should be exported and set
  to false on pure ipv4 nodes because of the globus bug

* DIRAC_GFAL_GRIDFTP_SESSION_REUSE: This should be exported
  and set to true in server bashrc files for efficiency reasons.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# pylint: disable=arguments-differ

# # imports
from past.builtins import long
import six
import os
import datetime
import errno
from stat import S_ISREG, S_ISDIR, S_IXUSR, S_IRUSR, S_IWUSR, S_IRWXG, S_IRWXU, S_IRWXO

import gfal2  # pylint: disable=import-error

# # from DIRAC
from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities import DErrno
from DIRAC.Resources.Storage.Utilities import checkArgumentFormat
from DIRAC.Resources.Storage.StorageBase import StorageBase
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.Core.Utilities.File import getSize
from DIRAC.Core.Utilities.Pfn import pfnparse, pfnunparse


# # RCSID
__RCSID__ = "$Id$"


class GFAL2_StorageBase(StorageBase):
    """.. class:: GFAL2_StorageBase

    SRM v2 interface to StorageElement using gfal2
    """

    def __init__(self, storageName, parameters):
        """c'tor

        :param self: self reference
        :param str storageName: SE name
        :param dict parameters: storage parameters
        """

        StorageBase.__init__(self, storageName, parameters)

        self.log = gLogger.getSubLogger("GFAL2_StorageBase")

        # Some storages have problem to compute the checksum with xroot while getting the files
        # This allows to disable the checksum calculation while transferring, and then we compare the
        # file size
        self.disableTransferChecksum = True if (parameters.get("DisableChecksum") == "True") else False

        # Different levels or verbosity:
        # gfal2.verbose_level.normal,
        # gfal2.verbose_level.verbose,
        # gfal2.verbose_level.debug,
        # gfal2.verbose_level.trace

        dlevel = self.log.getLevel()
        if dlevel == "DEBUG":
            gLogger.enableLogsFromExternalLibs()
            gfal2.set_verbose(gfal2.verbose_level.trace)

        self.isok = True

        # # gfal2 API
        self.ctx = gfal2.creat_context()

        # by default turn off BDII checks
        self.ctx.set_opt_boolean("BDII", "ENABLE", False)

        # session reuse should only be done on servers
        self.ctx.set_opt_boolean(
            "GRIDFTP PLUGIN",
            "SESSION_REUSE",
            os.environ.get("DIRAC_GFAL_GRIDFTP_SESSION_REUSE", "no").lower() in ["true", "yes"],
        )

        # Enable IPV6 for gsiftp
        self.ctx.set_opt_boolean(
            "GRIDFTP PLUGIN",
            "IPV6",
            os.environ.get("DIRAC_GFAL_GRIDFTP_ENABLE_IPV6", "true").lower() not in ["false", "no"],
        )

        # spaceToken used for copying from and to the storage element
        self.spaceToken = parameters.get("SpaceToken", "")
        # stageTimeout, default timeout to try and stage/pin a file
        self.stageTimeout = gConfig.getValue("/Resources/StorageElements/StageTimeout", 12 * 60 * 60)
        # gfal2Timeout, amount of time it takes until an operation times out
        self.gfal2Timeout = gConfig.getValue("/Resources/StorageElements/GFAL_Timeout", 100)
        # set the gfal2 default protocols, e.g. used when trying to retrieve transport url
        self.defaultLocalProtocols = gConfig.getValue("/Resources/StorageElements/DefaultProtocols", [])

        # # set checksum type, by default this is 0 (GFAL_CKSM_NONE)
        self.checksumType = gConfig.getValue("/Resources/StorageElements/ChecksumType", "0")

        if self.checksumType == "0":
            self.checksumType = None

        self.log.debug("GFAL2_StorageBase: using %s checksum" % self.checksumType)

        self.voName = None
        ret = getProxyInfo(disableVOMS=True)
        if ret["OK"] and "group" in ret["Value"]:
            self.voName = getVOForGroup(ret["Value"]["group"])

        self.MAX_SINGLE_STREAM_SIZE = 1024 * 1024 * 10  # 10 MB ???
        self.MIN_BANDWIDTH = 0.5 * (1024 * 1024)  # 0.5 MB/s ???

        # This is the list of extended metadata to query the server for.
        # It is used by getSingleMetadata.
        # If set to None, No extended metadata will be queried
        # If the list is empty, all of them will be queried
        self._defaultExtendedAttributes = []

    def exists(self, path):
        """Check if the path exists on the storage

        :param self: self reference
        :param str path: path or list of paths to be checked
        :returns: Failed dictionary: {pfn : error message}
                  Successful dictionary: {pfn : bool}
                  S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.exists: Checking the existence of %s path(s)" % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            res = self.__singleExists(url)

            if res["OK"]:
                successful[url] = res["Value"]
            else:  # something went wrong with the query
                failed[url] = res["Message"]

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def _estimateTransferTimeout(self, fileSize):
        """Dark magic to estimate the timeout for a transfer
        The values are set empirically and seem to work fine.
        They were evaluated with gfal1 and SRM.

        :param fileSize: size of the file in bytes we want to transfer

        :return: timeout in seconds
        """

        return int(fileSize / self.MIN_BANDWIDTH * 4 + 310)

    def __singleExists(self, path):
        """Check if :path: exists on the storage

        :param self: self reference
        :param str: path to be checked (srm://...)
        :returns: S_OK ( boolean exists ) a boolean whether it exists or not
                  S_ERROR( errStr ) there is a problem with getting the information
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._singleExists")
        log.debug("Determining whether %s exists or not" % path)

        try:
            self.ctx.stat(str(path))  # If path doesn't exist this will raise an error - otherwise path exists
            log.debug("path exists")
            return S_OK(True)
        except gfal2.GError as e:
            if e.code == errno.ENOENT:
                log.debug("Path does not exist")
                return S_OK(False)
            else:
                log.debug("GFAL2_StorageBase.__singleExists: %s" % repr(e))
                return S_ERROR(e.code, e.message)

    def isFile(self, path):
        """Check if the path provided is a file or not

        :param self: self reference
        :param str: path or list of paths to be checked ( 'srm://...')
        :returns: Failed dict: {path : error message}
                  Successful dict: {path : bool}
                  S_ERROR in case of argument problems

        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.isFile: checking whether %s path(s) are file(s)." % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            res = self._isSingleFile(url)

            if res["OK"]:
                successful[url] = res["Value"]
            else:
                failed[url] = res["Message"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _isSingleFile(self, path):
        """Checking if :path: exists and is a file

        :param self: self reference
        :param str path: single path on the storage (srm://...)

        :returns: S_ERROR if there is a fatal error
                  S_OK ( boolean) if it is a file or not
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._isSingleFile")
        log.debug("Determining whether %s is a file or not." % path)

        try:
            statInfo = self.ctx.stat(str(path))
            # instead of return S_OK( S_ISDIR( statInfo.st_mode ) ) we use if/else. So we can use the log.
            if S_ISREG(statInfo.st_mode):
                log.debug("Path is a file")
                return S_OK(True)
            else:
                log.debug("Path is not a file")
                return S_OK(False)
        except gfal2.GError as e:
            log.debug("GFAL2_StorageBase._isSingleFile: %s" % repr(e))
            return S_ERROR(e.code, e.message)

    def putFile(self, path, sourceSize=0):
        """Put a copy of a local file or a file on another srm storage to a directory on the
        physical storage.

        :param path: dictionary { lfn (srm://...) : localFile }
        :param sourceSize: size of the file in byte. Mandatory for third party copy (WHY ???)
                             Also, this parameter makes it essentially a non bulk operation for
                             third party copy, unless all files have the same size...
        :returns: Successful dict: { path : size }
                  Failed dict: { path : error message }
                  S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        failed = {}
        successful = {}

        for dest_url, src_file in urls.items():
            if not src_file:
                errStr = "GFAL2_StorageBase.putFile: Source file not set. Argument must be a dictionary \
                                             (or a list of a dictionary) {url : local path}"
                self.log.debug(errStr)
                return S_ERROR(errStr)
            res = self._putSingleFile(src_file, dest_url, sourceSize)

            if res["OK"]:
                successful[dest_url] = res["Value"]
            else:
                failed[dest_url] = res["Message"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _putSingleFile(self, src_file, dest_url, sourceSize):
        """Put a copy of the local file to the current directory on the
        physical storage

        For gfal2 version 2.7.8 and lower the environment variable GLOBUS_THREAD_MODEL has to be
        set to 'pthread' otherwise dirac-dms-add-file will go in a deadlock. This is fixed with gfal2 2.8.

        :param str src_file: local file to copy
        :param str dest_file: pfn (srm://...)
        :param int sourceSize: size of the source file
        :returns: S_OK( fileSize ) if everything went fine, S_ERROR otherwise
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._putSingleFile")
        log.debug("trying to upload %s to %s" % (src_file, dest_url))

        # in case src_file is not set (this is also done in putFile but some methods directly access this method,
        # so that's why it's checked here one more time
        if not src_file:
            errStr = "no source defined, please check argument format { destination : localfile }"
            return S_ERROR(errStr)

        # check whether the source is local or on another storage
        if any(src_file.startswith(protocol + ":") for protocol in self.protocolParameters["InputProtocols"]):
            src_url = src_file
            if not sourceSize:
                errStr = "For file replication the source file size in bytes must be provided."
                log.debug(errStr, src_file)
                return S_ERROR(errno.EINVAL, errStr)

        # file is local so we can set the protocol and determine source size accordingly
        else:
            if not os.path.isfile(src_file):
                errStr = "The local source file does not exist or is a directory"
                log.debug(errStr, src_file)
                return S_ERROR(errno.ENOENT, errStr)
            if not src_file.startswith("file://"):
                src_url = "file://%s" % os.path.abspath(src_file)
            sourceSize = getSize(src_file)
            if sourceSize == -1:
                errStr = "Failed to get file size"
                log.debug(errStr, src_file)
                return S_ERROR(DErrno.EFILESIZE, errStr)
            if sourceSize == 0:
                errStr = "Source file size is zero."
                log.debug(errStr, src_file)
                return S_ERROR(DErrno.EFILESIZE, errStr)

        # source is OK, creating the destination folder
        dest_path = os.path.dirname(dest_url)
        res = self._createSingleDirectory(dest_path)
        if not res["OK"]:
            log.debug("Failed to create destination folder %s" % dest_path, res["Message"])
            return res

        # folder is created and file exists, setting known copy parameters
        params = self.ctx.transfer_parameters()
        params.timeout = self._estimateTransferTimeout(sourceSize)
        if sourceSize > self.MAX_SINGLE_STREAM_SIZE:
            params.nbstreams = 4
        else:
            params.nbstreams = 1
        params.overwrite = (
            True  # old gfal removed old file first, gfal2 can just overwrite it with this flag set to True
        )
        if self.spaceToken:
            params.dst_spacetoken = self.spaceToken

        if self.checksumType:
            params.set_checksum(gfal2.checksum_mode.both, self.checksumType, "")

        # Params set, copying file now
        try:
            self.ctx.filecopy(params, str(src_url), str(dest_url))
            if self.checksumType:
                # checksum check is done by gfal2
                return S_OK(sourceSize)
            # no checksum check, compare file sizes for verfication
            else:
                res = self._getSingleFileSize(dest_url)
                # In case of failure, we set destSize to None
                # so that the cleaning of the file happens
                if not res["OK"]:
                    destSize = None
                else:
                    destSize = res["Value"]

                log.debug("destSize: %s, sourceSize: %s" % (destSize, sourceSize))
                if destSize == sourceSize:
                    return S_OK(destSize)
                else:
                    log.debug(
                        "Source and destination file size don't match.\
                                                                        Trying to remove destination file"
                    )
                    res = self._removeSingleFile(dest_url)
                    if not res["OK"]:
                        log.debug("Failed to remove destination file", res["Message"])
                        return res
                    errStr = "Source and destination file size don't match. Removed destination file"
                    log.debug(errStr, {sourceSize: destSize})
                    return S_ERROR("%s srcSize: %s destSize: %s" % (errStr, sourceSize, destSize))
        except gfal2.GError as e:
            # ##
            # extended error message because otherwise we could only guess what the error could be when we copy
            # from another srm to our srm-SE '''
            errStr = "Exception while copying"
            detailMsg = "Failed to copy file %s to destination url %s: [%d] %s" % (
                src_file,
                dest_url,
                e.code,
                e.message,
            )
            log.debug(errStr, detailMsg)
            return S_ERROR(e.code, detailMsg)

    def getFile(self, path, localPath=False):
        """Make a local copy of storage :path:

        :param self: self reference
        :param str path: path (or list of paths) on storage (srm://...)
        :param localPath: destination folder. Default is from current directory
        :returns: Successful dict: {path : size}
                  Failed dict: {path : errorMessage}
                  S_ERROR in case of argument problems
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.getFile: Trying to download %s files." % len(urls))

        failed = {}
        successful = {}

        for src_url in urls:
            fileName = os.path.basename(src_url)
            dest_file = os.path.join(localPath if localPath else os.getcwd(), fileName)

            res = self._getSingleFile(src_url, dest_file, disableChecksum=self.disableTransferChecksum)

            if not res["OK"]:
                failed[src_url] = res["Message"]
            else:
                successful[src_url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleFile(self, src_url, dest_file, disableChecksum=False):
        """Copy a storage file :src_url: to a local fs under :dest_file:

        :param self: self reference
        :param str src_url: SE url that is to be copied (srm://...)
        :param str dest_file: local fs path
        :param bool disableChecksum: There are problems with xroot comparing checksums after
                                     copying a file so with this parameter we can disable checksum
                                     checks for xroot
        :returns: S_ERROR( errStr ) in case of an error
                  S_OK( size of file ) if copying is successful
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._getSingleFile")

        log.info("Trying to download %s to %s" % (src_url, dest_file))
        if disableChecksum:
            log.warn("checksum calculation disabled for transfers!")

        destDir = os.path.dirname(dest_file)

        if not os.path.exists(destDir):
            log.debug("Local directory does not yet exist. Creating it", destDir)
            try:
                os.makedirs(destDir)
            except OSError as error:
                errStr = "Error while creating the destination folder"
                log.exception(errStr, lException=error)
                return S_ERROR("%s: %s" % (errStr, repr(error)))

        res = self._getSingleFileSize(src_url)
        if not res["OK"]:
            log.debug("Error while determining file size", res["Message"])
            return res

        remoteSize = res["Value"]

        # Set gfal2 copy parameters
        # folder is created and file exists, setting known copy parameters
        params = self.ctx.transfer_parameters()
        params.timeout = self._estimateTransferTimeout(remoteSize)
        if remoteSize > self.MAX_SINGLE_STREAM_SIZE:
            params.nbstreams = 4
        else:
            params.nbstreams = 1
        params.overwrite = (
            True  # old gfal removed old file first, gfal2 can just overwrite it with this flag set to True
        )
        if self.spaceToken:
            params.src_spacetoken = self.spaceToken

        useChecksum = bool(self.checksumType and not disableChecksum)
        if useChecksum:
            # params.set_user_defined_checksum(self.checksumType, '')
            params.set_checksum(gfal2.checksum_mode.both, self.checksumType, "")

        # Params set, copying file now
        try:
            # gfal2 needs a protocol to copy local which is 'file:'
            if not dest_file.startswith("file://"):
                dest = "file://%s" % os.path.abspath(dest_file)
            self.ctx.filecopy(params, str(src_url), str(dest))
            if useChecksum:
                # gfal2 did a checksum check, so we should be good
                return S_OK(remoteSize)
            else:
                # No checksum check was done so we compare file sizes
                destSize = getSize(dest_file)
                if destSize == remoteSize:
                    return S_OK(destSize)
                else:
                    errStr = "File sizes don't match. Something went wrong. Removing local file %s" % dest_file
                    log.debug(errStr, {remoteSize: destSize})
                    if os.path.exists(dest_file):
                        os.remove(dest_file)
                    return S_ERROR(errStr)
        except gfal2.GError as e:
            errStr = "Could not copy %s to %s, [%d] %s" % (src_url, dest, e.code, e.message)
            log.debug(errStr)
            return S_ERROR(e.code, errStr)

    def removeFile(self, path):
        """Physically remove the file specified by path

        A non existing file will be considered as successfully removed

        :param str path: path (or list of paths) on storage (srm://...)
        :returns: Successful dict {path : True}
                   Failed dict {path : error message}
                   S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.removeFile: Attempting to remove %s files" % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            res = self._removeSingleFile(url)

            if res["OK"]:
                successful[url] = res["Value"]
            else:
                failed[url] = res["Message"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _removeSingleFile(self, path):
        """Physically remove the file specified by path

        :param str path: path on storage (srm://...)
        :returns: S_OK( True )  if the removal was successful (also if file didn't exist in the first place)
                  S_ERROR( errStr ) if there was a problem removing the file
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._removeSingleFile")
        log.debug("Attempting to remove single file %s" % path)
        path = str(path)
        try:
            status = self.ctx.unlink(str(path))
            if status == 0:
                log.debug("File successfully removed")
                return S_OK(True)
            elif status < 0:
                errStr = "return status < 0. Error occurred."
                return S_ERROR(errStr)
        except gfal2.GError as e:
            # file doesn't exist so operation was successful
            if e.code == errno.ENOENT:
                log.debug("File does not exist.")
                return S_OK(True)
            elif e.code == errno.EISDIR:
                log.debug("Path is a directory.")
                return S_ERROR(errno.EISDIR, errStr)
            else:
                errStr = "Failed to remove file."
                log.debug("Failed to remove file: [%d] %s" % (e.code, e.message))
                return S_ERROR(e.code, repr(e))

    def getFileSize(self, path):
        """Get the physical size of the given file

        :param self: self reference
        :param path: path (or list of path) on storage (pfn : srm://...)
        :returns: Successful dict {path : size}
               Failed dict {path : error message }
               S_ERROR in case of argument problems
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.getFileSize: Trying to determine file size of %s files" % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            res = self._getSingleFileSize(url)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleFileSize(self, path):
        """Get the physical size of the given file

        :param self: self reference
        :param path: single path on the storage (srm://...)
        :returns: S_OK( filesize ) when successfully determined filesize
                  S_ERROR( errStr ) filesize could not be determined
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._getSingleFileSize")
        log.debug("Determining file size of %s" % path)
        path = str(path)
        try:
            statInfo = self.ctx.stat(str(path))  # keeps info like size, mode.

            # If it is not a file
            if not S_ISREG(statInfo.st_mode):
                errStr = "Path is not a file"
                self.log.debug(errStr)
                return S_ERROR(errno.EISDIR, errStr)

            self.log.debug("File size successfully determined %s" % statInfo.st_size)
            return S_OK(long(statInfo.st_size))
        except gfal2.GError as e:
            errStr = "Failed to determine file size."
            self.log.debug(errStr, repr(e))
            return S_ERROR(e.code, "%s: %s" % (errStr, repr(e)))

    def getFileMetadata(self, path):
        """Get metadata associated to the file(s)

        :param self: self reference
        :param str path: path (or list of paths) on the storage (srm://...)
        :returns: successful dict { path : metadata }
                 failed dict { path : error message }
                 S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.getFileMetadata: trying to read metadata for %s paths" % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            res = self._getSingleFileMetadata(url)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleFileMetadata(self, path):
        """Fetch the metadata associated to the file
        :param self: self reference
        :param path: path (only 1) on storage (srm://...)
        :returns:
            S_OK (MetadataDict) if we could get the metadata
            S_ERROR (errorMsg) if there was a problem getting the metadata or if it is not a file
        """
        self.log.debug("GFAL2_StorageBase._getSingleFileMetadata: trying to read metadata for %s" % path)

        res = self._getSingleMetadata(path)

        if not res["OK"]:
            return res

        metaDict = res["Value"]

        if not metaDict["File"]:
            errStr = "GFAL2_StorageBase._getSingleFileMetadata: supplied path is not a file"
            self.log.debug(errStr, path)
            return S_ERROR(errno.EISDIR, errStr)

        return S_OK(metaDict)

    def _updateMetadataDict(self, _metadataDict, _attributeDict):
        """Updating the metadata dictionary with protocol specific attributes
          Dummy implementation
        :param self: self reference
        :param dict: metadataDict we want add the specific attributes to
        :param dict: attributeDict contains the special attributes

        """

        return

    def _getSingleMetadata(self, path):
        """Fetches the metadata of a single file or directory via gfal2.stat
           and getExtendedAttributes

        :param self: self reference
        :param path: path (only 1) on storage (srm://...)
        :returns:
            S_OK ( MetadataDict ) if we could get the metadata
            S_ERROR ( errorMsg ) if there was a problem getting the metadata
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._getSingleMetadata")
        log.debug("Reading metadata for %s" % path)

        try:
            statInfo = self.ctx.stat(str(path))
        except gfal2.GError as e:
            errStr = "Failed to retrieve metadata"
            self.log.debug(errStr, repr(e))
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))

        metadataDict = self.__parseStatInfoFromApiOutput(statInfo)

        if metadataDict["File"] and self.checksumType:
            res = self.__getChecksum(path, self.checksumType)
            if not res["OK"]:
                log.warn("Could not get checksum:%s" % res["Message"])
            metadataDict["Checksum"] = res.get("Value", "")

        metadataDict = self._addCommonMetadata(metadataDict)

        if self._defaultExtendedAttributes is not None:
            res = self._getExtendedAttributes(path, attributes=self._defaultExtendedAttributes)
            if not res["OK"]:
                log.warn("Could not get extended attributes: %s" % res["Message"])
            else:
                attributeDict = res["Value"]
                # add extended attributes to the dict if available
                self._updateMetadataDict(metadataDict, attributeDict)

        return S_OK(metadataDict)

    def prestageFile(self, path, lifetime=86400):
        """Issue prestage request for file(s)

        :param self: self reference
        :param str path: path or list of paths to be prestaged
        :param int lifetime: prestage lifetime in seconds (default 24h)

        :return: succesful dict { url : token }
                failed dict { url : message }
                S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.prestageFile: Attempting to issue stage requests for %s file(s)." % len(urls))

        failed = {}
        successful = {}
        for url in urls:
            res = self._prestageSingleFile(url, lifetime)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]
        return S_OK({"Failed": failed, "Successful": successful})

    def _prestageSingleFile(self, path, lifetime):
        """Issue prestage for single file

        :param self: self reference
        :param str path: path to be prestaged
        :param int lifetime: prestage lifetime in seconds (default 24h)

        :return: S_ structure
                                S_OK( token ) ) if status >= 0 (0 - staging is pending, 1 - file is pinned)
                                S_ERROR( errMsg ) ) in case of an error: status -1
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._prestageSingleFile")
        log.debug("Attempting to issue stage request for single file: %s" % path)

        try:
            (status, token) = self.ctx.bring_online(str(path), lifetime, self.stageTimeout, True)
            log.debug("Staging issued - Status: %s" % status)
            if status >= 0:
                return S_OK(token)
            else:
                return S_ERROR("An error occured while issuing prestaging.")
        except gfal2.GError as e:
            errStr = "Error occured while prestaging file"
            log.debug(errStr, "%s %s" % (path, repr(e)))
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))

    def prestageFileStatus(self, path):
        """Checking the staging status of file(s) on the storage

        :param self: self reference
        :param dict path: dict { url : token }
        :return: succesful dict { url : bool }
                failed dict { url : message }
                S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.prestageFileStatus: Checking the staging status for %s file(s)." % len(urls))

        failed = {}
        successful = {}
        for path, token in urls.items():

            res = self._prestageSingleFileStatus(path, token)
            if not res["OK"]:
                failed[path] = res["Message"]
            else:
                successful[path] = res["Value"]
        return S_OK({"Failed": failed, "Successful": successful})

    def _prestageSingleFileStatus(self, path, token):
        """Check prestage status for single file

        :param self: self reference
        :param str path: path to be checked
        :param str token: token of the file

        :return: S_ structure
                                S_OK( True ) if file is staged
                                S_OK( False ) if file is not staged yet
                                S_ERROR( errMsg ) ) in case of an error: status -1
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._prestageSingleFileStatus")
        log.debug("Checking prestage file status for %s" % path)
        # also allow int as token - converting them to strings
        if not isinstance(token, six.string_types):
            token = str(token)

        try:
            self.ctx.set_opt_boolean("BDII", "ENABLE", True)
            status = self.ctx.bring_online_poll(str(path), str(token))

            if status == 0:
                log.debug("File not staged")
                return S_OK(False)
            elif status == 1:
                log.debug("File is staged")
                return S_OK(True)
            else:
                return S_ERROR("An error occured while checking prestage status.")
        except gfal2.GError as e:
            if e.code == errno.EAGAIN:
                log.debug("File not staged")
                return S_OK(False)
            elif e.code == errno.ETIMEDOUT:
                errStr = "Polling request timed out"
                log.debug(errStr)
                return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))
            else:
                errStr = "Error occured while polling for prestaging file"
                log.debug(errStr, "%s %s" % (path, repr(e)))
                return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))
        finally:
            self.ctx.set_opt_boolean("BDII", "ENABLE", False)

    def pinFile(self, path, lifetime=86400):
        """Pin a staged file

        :param self: self reference
        :param str path: path of list of paths to be pinned
        :param int lifetime: pinning time in seconds (default 24h)

        :return: successful dict {url : token},
                 failed dict {url : message}
                 S_ERROR in case of argument problems
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.pinFile: Attempting to pin %s file(s)." % len(urls))
        failed = {}
        successful = {}
        for url in urls:
            res = self._pinSingleFile(url, lifetime)
            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]
        return S_OK({"Failed": failed, "Successful": successful})

    def _pinSingleFile(self, path, lifetime):
        """Pin a single staged file

        :param self: self reference
        :param str path: path to be pinned
        :param int lifetime: pinning lifetime in seconds (default 24h)

        :return:  S_OK( token ) ) if status >= 0 (0 - staging is pending, 1 - file is pinned). EAGAIN is also considered pending
                  S_ERROR( errMsg ) ) in case of an error: status -1
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._pinSingleFile")
        log.debug("Attempting to issue pinning request for single file: %s" % path)

        try:
            self.ctx.set_opt_boolean("BDII", "ENABLE", True)
            status, token = self.ctx.bring_online(str(path), lifetime, self.stageTimeout, True)
            log.debug("Pinning issued - Status: %s" % status)
            if status >= 0:
                return S_OK(token)
            else:
                return S_ERROR("An error occured while issuing pinning.")
        except gfal2.GError as e:
            errStr = "GFAL2_StorageBase._pinSingleFile: Error occured while pinning file"
            log.debug(errStr, "%s %s" % (path, repr(e)))
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))
        finally:
            self.ctx.set_opt_boolean("BDII", "ENABLE", False)

    def releaseFile(self, path):
        """Release a pinned file

        :param self: self reference
        :param str path: PFN path { pfn : token } - pfn can be an empty string, then all files that have that same token get released.
                         Just as you can pass an empty token string and a directory as pfn which then releases all the files in the directory
                         an its subdirectories

        :return: successful dict {url : token},
                 failed dict {url : message}
                 S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.releaseFile: Attempting to release %s file(s)." % len(urls))

        failed = {}
        successful = {}
        for path, token in urls.items():
            res = self._releaseSingleFile(path, token)

            if not res["OK"]:
                failed[path] = res["Message"]
            else:
                successful[path] = res["Value"]
        return S_OK({"Failed": failed, "Successful": successful})

    def _releaseSingleFile(self, path, token):
        """release a single pinned file

        :param self: self reference
        :param str path: path to the file to be released
        :token str token: token belonging to the path

        :returns: S_OK( token ) when releasing was successful, S_ERROR( errMessage ) in case of an error
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._releaseSingleFile")
        log.debug("Attempting to release single file: %s" % path)
        if not isinstance(token, six.string_types):
            token = str(token)
        try:
            self.ctx.set_opt_boolean("BDII", "ENABLE", True)
            status = self.ctx.release(str(path), token)
            if status >= 0:
                return S_OK(token)
            else:
                errStr = "Error occured: Return status < 0"
                log.debug(errStr, "path %s token %s" % (path, token))
                return S_ERROR(errStr)
        except gfal2.GError as e:
            errStr = "Error occured while releasing file"
            self.log.debug(errStr, "%s %s" % (path, repr(e)))
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))
        finally:
            self.ctx.set_opt_boolean("BDII", "ENABLE", False)

    def __getChecksum(self, path, checksumType=None):
        """Calculate the checksum (ADLER32 by default) of a file on the storage

        :param self: self reference
        :param str path: path to single file on storage (srm://...)
        :returns: S_OK( checksum ) if checksum could be calculated
                  S_ERROR( errMsg ) if something failed
        """
        log = self.log.getSubLogger(GFAL2_StorageBase.__getChecksum)
        log.debug("Trying to calculate checksum of file %s" % path)

        if not checksumType:
            errStr = "No checksum type set by the storage element. Can't retrieve checksum"
            log.debug(errStr, path)
            return S_ERROR(errStr)

        res = self._isSingleFile(path)
        if not res["OK"]:
            return res

        try:
            log.debug("using %s checksum" % checksumType)
            fileChecksum = self.ctx.checksum(str(path), checksumType)
            return S_OK(fileChecksum)

        except gfal2.GError as e:
            errStr = "Failed to calculate checksum."
            log.debug(errStr, repr(e))
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))

    def __parseStatInfoFromApiOutput(self, statInfo):
        """Fill the metaDict with the information obtained with gfal2.stat()

        returns metaDict with following keys:

        st_dev: ID of device containing file
        st_ino: file serial number
        st_mode: mode of file
        st_nlink: number of links to the file
        st_uid: user ID of file
        st_gid: group ID of file
        st_size: file size in bytes
        st_atime: time of last access
        st_mtime: time of last modification
        st_ctime: time of last status chage
        File (bool): whether object is a file or not
        Directory (bool): whether object is a directory or not
        """
        metaDict = {}
        # to identify whether statInfo are from file or directory
        metaDict["File"] = S_ISREG(statInfo.st_mode)
        metaDict["Directory"] = S_ISDIR(statInfo.st_mode)

        if metaDict["File"]:
            metaDict["FileSerialNumber"] = statInfo.st_ino
            metaDict["Mode"] = statInfo.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)
            metaDict["Links"] = statInfo.st_nlink
            metaDict["UserID"] = statInfo.st_uid
            metaDict["GroupID"] = statInfo.st_gid
            metaDict["Size"] = long(statInfo.st_size)
            metaDict["LastAccess"] = self.__convertTime(statInfo.st_atime) if statInfo.st_atime else "Never"
            metaDict["ModTime"] = self.__convertTime(statInfo.st_mtime) if statInfo.st_mtime else "Never"
            metaDict["StatusChange"] = self.__convertTime(statInfo.st_ctime) if statInfo.st_ctime else "Never"
            metaDict["Executable"] = bool(statInfo.st_mode & S_IXUSR)
            metaDict["Readable"] = bool(statInfo.st_mode & S_IRUSR)
            metaDict["Writeable"] = bool(statInfo.st_mode & S_IWUSR)
        elif metaDict["Directory"]:
            metaDict["Mode"] = statInfo.st_mode & (S_IRWXU | S_IRWXG | S_IRWXO)

        return metaDict

    @staticmethod
    def __convertTime(time):
        """Converts unix time to proper time format

        :param self: self reference
        :param time: unix time
        :return: Date in following format: 2014-10-29 14:32:10
        """
        return datetime.datetime.fromtimestamp(time).strftime("%Y-%m-%d %H:%M:%S")

    def createDirectory(self, path):
        """Create directory on the storage

        :param self: self reference
        :param str path: path to be created on the storage (pfn : srm://...)
        :returns: Successful dict {path : True }
                 Failed dict     {path : error message }
                 S_ERROR in case of argument problems
        """
        urls = checkArgumentFormat(path)
        if not urls["OK"]:
            return urls
        urls = urls["Value"]

        successful = {}
        failed = {}
        self.log.debug("createDirectory: Attempting to create %s directories." % len(urls))
        for url in urls:
            res = self._createSingleDirectory(url)
            if res["OK"]:
                successful[url] = True
            else:
                failed[url] = res["Message"]
        return S_OK({"Failed": failed, "Successful": successful})

    def _createSingleDirectory(self, path):
        """Create directory :path: on the storage
        if no exception is caught the creation was successful. Also if the
        directory already exists we return S_OK().

        :param self: self reference
        :param str path: path to be created (srm://...)

        :returns: S_OK() if creation was successful or directory already exists
                 S_ERROR() in case of an error during creation
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._createSingleDirectory")
        try:
            log.debug("Creating %s" % path)
            status = self.ctx.mkdir_rec(str(path), 755)
            if status >= 0:
                log.debug("Successfully created directory")
                return S_OK()
            else:
                errStr = "Failled to create directory. Status return > 0."
                log.debug(errStr, status)
                return S_ERROR(errStr)
        except gfal2.GError as e:
            # error: directory already exists
            # Explanations for ECOMM:
            # Because of the way the new DPM DOME flavor works
            # and the poor error handling of Globus works, we might
            # encounter ECOMM when creating an existing directory
            # This will be fixed in the future versions of DPM,
            # but in the meantime, we catch it ourselves.
            if e.code in (errno.EEXIST, errno.ECOMM):
                log.debug("Directory already exists")
                return S_OK()
            # any other error: failed to create directory
            else:
                errStr = "Failed to create directory."
                log.debug(errStr, repr(e))
                return S_ERROR(e.code, repr(e))

    def isDirectory(self, path):
        """check if the path provided is a directory or not

        :param self: self reference
        :param str: path or list of paths to be checked ( 'srm://...')
        :returns: dict 'Failed' : failed, 'Successful' : succesful
                 S_ERROR in case of argument problems

        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.isDirectory: checking whether %s path(s) are directory(ies)." % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            res = self._isSingleDirectory(url)
            if res["OK"]:
                successful[url] = res["Value"]
            else:
                failed[url] = res["Message"]

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def _isSingleDirectory(self, path):
        """Checking if :path: exists and is a directory

        :param self: self reference
        :param str path: single path on the storage (srm://...)

        :returns: S_OK ( boolean) if it is a directory or not
                  S_ERROR ( errStr ) when there was a problem getting the info
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._isSingleDirectory")
        log.debug("Determining whether %s is a directory or not." % path)
        try:
            statInfo = self.ctx.stat(str(path))
            # instead of return S_OK( S_ISDIR( statInfo.st_mode ) ) we use if/else. So we can use the log.
            if S_ISDIR(statInfo.st_mode):
                log.debug("Path is a directory")
                return S_OK(True)
            else:
                log.debug("Path is not a directory")
                return S_OK(False)
        except gfal2.GError as e:
            errStr = "Failed to determine if path %s is a directory." % path
            log.debug(errStr, repr(e))
            return S_ERROR(e.code, repr(e))

    def listDirectory(self, path):
        """List the content of the path provided

        :param str path: single or list of paths (srm://...)
        :return: failed  dict {path : message }
                successful dict { path :  {'SubDirs' : subDirs, 'Files' : files} }.
                They keys are the paths, the values are the dictionary 'SubDirs' and 'Files'.
                Each are dictionaries with path as key and metadata as values
                S_ERROR in case of argument problems
        """

        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.listDirectory: Attempting to list %s directories" % len(urls))
        res = self.isDirectory(urls)
        if not res["OK"]:
            return res
        successful = {}
        failed = res["Value"]["Failed"]

        directories = []

        for url, isDirectory in res["Value"]["Successful"].items():
            if isDirectory:
                directories.append(url)
            else:
                errStr = "GFAL2_StorageBase.listDirectory: path is not a directory"
                gLogger.error(errStr, url)
                failed[url] = errStr

        for directory in directories:
            res = self._listSingleDirectory(directory)
            if not res["OK"]:
                failed[directory] = res["Message"]
            else:
                successful[directory] = res["Value"]

        resDict = {"Failed": failed, "Successful": successful}
        return S_OK(resDict)

    def _listSingleDirectory(self, path, internalCall=False):
        """List the content of the single directory provided

        :param self: self reference
        :param str path: single path on storage (srm://...)
        :param bool internalCall: if we call this method from another internal method we want
                                  to work with the full pfn. Used for _getSingleDirectory and
                                  _removeSingleDirectory
        :returns: S_ERROR( errStr ) if there is an error
                 S_OK( dictionary ): Key: SubDirs and Files
                                     The values of the Files are dictionaries with filename as key and metadata as value
                                     The values of SubDirs are just the dirnames as key and True as value
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._listSingleDirectory")
        log.debug("Attempting to list content of %s" % path)

        try:
            listing = self.ctx.listdir(str(path))

        except gfal2.GError as e:
            errStr = "Could not list directory content."
            log.debug(errStr, e.message)
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))

        files = {}
        subDirs = {}

        res = pfnparse(path, srmSpecific=self.srmSpecificParse)
        if not res["OK"]:
            return res
        pathDict = res["Value"]

        for entry in listing:

            nextEntry = dict(pathDict)
            nextEntry["FileName"] = os.path.join(pathDict["FileName"], entry)
            res = pfnunparse(nextEntry, srmSpecific=self.srmSpecificParse)
            if not res["OK"]:
                log.debug("Cannot generate url for next entry", res)
                continue

            nextUrl = res["Value"]

            res = self._getSingleMetadata(nextUrl)
            if res["OK"]:
                metadataDict = res["Value"]
                if internalCall:
                    subPathLFN = nextUrl
                else:
                    # If it is not an internal call, we return the LFN
                    # We cannot use a simple replace because of the double slash
                    # that might be at the start
                    basePath = os.path.normpath(self.protocolParameters["Path"])
                    startBase = nextEntry["Path"].find(basePath)
                    lfnStart = nextEntry["Path"][startBase + len(basePath) :]
                    if not lfnStart:
                        lfnStart = "/"
                    subPathLFN = os.path.join(lfnStart, nextEntry["FileName"])

                if metadataDict["Directory"]:
                    subDirs[subPathLFN] = metadataDict
                elif metadataDict["File"]:
                    files[subPathLFN] = metadataDict
                else:
                    log.debug("Found item which is neither file nor directory", nextUrl)
            else:
                log.debug("Could not stat content", "%s %s" % (nextUrl, res["Message"]))

        return S_OK({"SubDirs": subDirs, "Files": files})

    def getDirectory(self, path, localPath=False):
        """get a directory from the SE to a local path with all its files and subdirectories

        :param str path: path (or list of paths) on the storage (srm://...)
        :param str localPath: local path where the content of the remote directory will be saved,
                                if not defined it takes current working directory.
        :return: successful and failed dictionaries. The keys are the paths,
                 the values are dictionary {'Files': amount of files downloaded, 'Size' : amount of data downloaded}
                 S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        log = self.log.getSubLogger("GFAL2_StorageBase.getDirectory")
        log.debug("Attempting to get local copies of %s directories. %s" % (len(urls), urls))

        failed = {}
        successful = {}

        for src_dir in urls:
            res = pfnparse(src_dir, srmSpecific=self.srmSpecificParse)
            if not res["OK"]:
                log.debug("cannot parse src_url", res)
                continue
            srcUrlDict = res["Value"]
            dirName = srcUrlDict["FileName"]

            dest_dir = os.path.join(localPath if localPath else os.getcwd(), dirName)

            res = self._getSingleDirectory(src_dir, dest_dir)

            if res["OK"]:
                if res["Value"]["AllGot"]:
                    log.debug("Successfully got local copy of %s" % src_dir)
                    successful[src_dir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
                else:
                    log.debug("Failed to get entire directory.", src_dir)
                    failed[src_dir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
            else:
                log.debug("Completely failed to get local copy of directory.", src_dir)
                failed[src_dir] = {"Files": 0, "Size": 0}

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleDirectory(self, src_dir, dest_dir):
        """Download a single directory recursively
        :param self: self reference
        :param src_dir : remote directory to download (srm://...)
        :param dest_dir: local destination path
        :returns: S_ERROR if there is a fatal error
                S_OK if we could download something :
                              'AllGot': boolean of whether we could download everything
                              'Files': amount of files received
                              'Size': amount of data received
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._getSingleDirectory")
        log.debug("Attempting to download directory %s at %s" % (src_dir, dest_dir))

        filesReceived = 0
        sizeReceived = 0

        res = self._isSingleDirectory(src_dir)
        if not res["OK"]:
            log.debug("Failed to find the source directory: %s %s" % (res["Message"], src_dir))
            return res

        # res['Value'] is False if it's not a directory
        if not res["Value"]:
            errStr = "The path provided is not a directory"
            log.debug(errStr, src_dir)
            return S_ERROR(errno.ENOTDIR, errStr)

        if not os.path.exists(dest_dir):
            try:
                os.makedirs(dest_dir)
            except OSError as error:
                errStr = "Error trying to create destination directory %s" % error
                log.exception(errStr, lException=error)
                return S_ERROR(errStr)

        # Get the remote directory contents
        res = self._listSingleDirectory(src_dir, internalCall=True)
        if not res["OK"]:
            return res

        sFilesDict = res["Value"]["Files"]
        subDirsDict = res["Value"]["SubDirs"]

        # Get all the files in the directory
        receivedAllFiles = True
        log.debug("Trying to download the %s files" % len(sFilesDict))
        for sFile in sFilesDict:
            # Getting the last filename
            res = pfnparse(sFile, srmSpecific=self.srmSpecificParse)
            if not res["OK"]:
                log.debug("Cannot unparse target file. Skipping", res)
                receivedAllFiles = False
                continue
            filename = res["Value"]["FileName"]
            # Returns S_OK(fileSize) if successful
            res = self._getSingleFile(
                sFile, os.path.join(dest_dir, filename), disableChecksum=self.disableTransferChecksum
            )
            if res["OK"]:
                filesReceived += 1
                sizeReceived += res["Value"]
            else:
                receivedAllFiles = False

        # recursion to get contents of sub directoryies
        receivedAllDirs = True
        log.debug("Trying to recursively download the %s directories" % len(subDirsDict))
        for subDir in subDirsDict:
            # Getting the last filename
            res = pfnparse(subDir, srmSpecific=self.srmSpecificParse)
            if not res["OK"]:
                log.debug("Cannot unparse target dir. Skipping", res)
                receivedAllDirs = False
                continue
            subDirName = res["Value"]["FileName"]
            localPath = os.path.join(dest_dir, subDirName)
            res = self._getSingleDirectory(subDir, localPath)

            if not res["OK"]:
                receivedAllDirs = False
            else:
                if not res["Value"]["AllGot"]:
                    receivedAllDirs = False
                filesReceived += res["Value"]["Files"]
                sizeReceived += res["Value"]["Size"]

        allGot = receivedAllDirs and receivedAllFiles

        resDict = {"AllGot": allGot, "Files": filesReceived, "Size": sizeReceived}
        return S_OK(resDict)

    def putDirectory(self, path):
        """Puts one or more local directories to the physical storage together with all its files

        :param self: self reference
        :param str path: dictionary { srm://... (destination) : localdir (source dir) }
        :return: successful and failed dictionaries. The keys are the paths,
                 the values are dictionary {'Files' : amount of files uploaded, 'Size' : amount of data upload }
                 S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        log = self.log.getSubLogger("GFAL2_StorageBase.putDirectory")

        log.debug("Attempting to put %s directories to remote storage" % len(urls))

        successful = {}
        failed = {}
        for destDir, sourceDir in urls.items():
            if not sourceDir:
                errStr = "No source directory set, make sure the input format is correct { dest. dir : source dir }"
                return S_ERROR(errStr)
            res = self._putSingleDirectory(sourceDir, destDir)
            if res["OK"]:
                if res["Value"]["AllPut"]:
                    log.debug("Successfully put directory to remote storage: %s" % destDir)
                    successful[destDir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
                else:
                    log.debug("Failed to put entire directory to remote storage.", destDir)
                    failed[destDir] = {"Files": res["Value"]["Files"], "Size": res["Value"]["Size"]}
            else:
                log.debug("Completely failed to put directory to remote storage.", destDir)
                failed[destDir] = {"Files": 0, "Size": 0}
        return S_OK({"Failed": failed, "Successful": successful})

    def _putSingleDirectory(self, src_directory, dest_directory):
        """puts one local directory to the physical storage together with all its files and subdirectories
        :param self: self reference
        :param src_directory : the local directory to copy
        :param dest_directory: pfn (srm://...) where to copy
        :returns: S_ERROR if there is a fatal error
                  S_OK if we could upload something :
                                    'AllPut': boolean of whether we could upload everything
                                    'Files': amount of files uploaded
                                    'Size': amount of data uploaded
        """
        log = self.log.getSubLogger("GFAL2_StorageBase._putSingleDirectory")
        log.debug("Trying to upload %s to %s" % (src_directory, dest_directory))
        filesPut = 0
        sizePut = 0

        if not os.path.isdir(src_directory):
            errStr = "The supplied source directory does not exist or is not a directory."
            log.debug(errStr, src_directory)
            return S_ERROR(errno.ENOENT, errStr)

        contents = os.listdir(src_directory)
        allSuccessful = True
        directoryFiles = {}

        res = pfnparse(dest_directory, srmSpecific=self.srmSpecificParse)
        if not res["OK"]:
            return res
        destDirParse = res["Value"]
        for fileName in contents:
            localPath = os.path.join(src_directory, fileName)

            nextUrlDict = dict(destDirParse)
            nextUrlDict["FileName"] = os.path.join(destDirParse["FileName"], fileName)
            res = pfnunparse(nextUrlDict, srmSpecific=self.srmSpecificParse)
            if not res["OK"]:
                log.debug("Cannot unparse next url dict. Skipping", res)
                allSuccessful = False
                continue

            remoteUrl = res["Value"]

            # if localPath is not a directory put it to the files dict that needs to be uploaded
            if not os.path.isdir(localPath):
                directoryFiles[remoteUrl] = localPath
            # localPath is another folder, start recursion
            else:
                res = self._putSingleDirectory(localPath, remoteUrl)
                if not res["OK"]:
                    log.debug("Failed to put directory to storage. Skipping", res["Message"])
                    allSuccessful = False
                else:
                    if not res["Value"]["AllPut"]:
                        allSuccessful = False
                    filesPut += res["Value"]["Files"]
                    sizePut += res["Value"]["Size"]

        if directoryFiles:
            res = self.putFile(directoryFiles)
            if not res["OK"]:
                log.debug("Failed to put files to storage.", res["Message"])
                allSuccessful = False
            else:
                for fileSize in res["Value"]["Successful"].values():
                    filesPut += 1
                    sizePut += fileSize
                if res["Value"]["Failed"]:
                    allSuccessful = False
        return S_OK({"AllPut": allSuccessful, "Files": filesPut, "Size": sizePut})

    def removeDirectory(self, path, recursive=False):
        """Remove a directory on the physical storage together with all its files and
        subdirectories.

        :param path: single or list of path (srm://..)
        :param recursive: if True, we recursively delete the subdir
        :return: successful and failed dictionaries. The keys are the pathes,
                  the values are dictionary {'Files': amount of files deleted, 'Size': amount of data deleted}
                  S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        log = self.log.getSubLogger("GFAL2_StorageBase.removeDirectory")
        log.debug("Attempting to remove %s directories." % len(urls))

        successful = {}
        failed = {}

        for url in urls:
            res = self._removeSingleDirectory(url, recursive)

            if res["OK"]:
                if res["Value"]["AllRemoved"]:
                    log.debug("Successfully removed %s" % url)
                    successful[url] = {
                        "FilesRemoved": res["Value"]["FilesRemoved"],
                        "SizeRemoved": res["Value"]["SizeRemoved"],
                    }
                else:
                    log.debug("Failed to remove entire directory.", path)
                    failed[url] = {
                        "FilesRemoved": res["Value"]["FilesRemoved"],
                        "SizeRemoved": res["Value"]["SizeRemoved"],
                    }
            else:
                log.debug("Completely failed to remove directory.", url)
                failed[url] = res["Message"]  # {'FilesRemoved':0, 'SizeRemoved':0}

        return S_OK({"Failed": failed, "Successful": successful})

    def _removeSingleDirectory(self, path, recursive=False):
        """Remove a directory on the physical storage together with all its files and
        subdirectories.
        :param path: pfn (srm://...) of a directory to remove
        :param recursive : if True, we recursively delete the subdir
        :returns: S_ERROR if there is a fatal error
                   S_OK (statistics dictionary ) if we could upload something :
                                     'AllRemoved': boolean of whether we could delete everything
                                     'FilesRemoved': amount of files deleted
                                     'SizeRemoved': amount of data deleted
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._removeSingleDirectory")
        filesRemoved = 0
        sizeRemoved = 0

        # Check the remote directory exists

        res = self._isSingleDirectory(path)

        if not res["OK"]:
            return res

        # res['Value'] is True if it is a directory
        if not res["Value"]:
            errStr = "The supplied path is not a directory."
            log.debug(errStr, path)
            return S_ERROR(errno.ENOTDIR, errStr)

        # Get the remote directory contents
        res = self._listSingleDirectory(path, internalCall=True)
        if not res["OK"]:
            return res

        sFilesDict = res["Value"]["Files"]
        subDirsDict = res["Value"]["SubDirs"]

        removedAllFiles = True
        removedAllDirs = True

        # if recursive, we call ourselves on all the subdirs
        if recursive:
            # Recursively remove the sub directories
            log.debug("Trying to recursively remove %s folder." % len(subDirsDict))
            for subDirUrl in subDirsDict:
                res = self._removeSingleDirectory(subDirUrl, recursive)
                if not res["OK"]:
                    log.debug("Recursive removal failed", res)
                    removedAllDirs = False
                else:
                    if not res["Value"]["AllRemoved"]:
                        removedAllDirs = False
                    filesRemoved += res["Value"]["FilesRemoved"]
                    sizeRemoved += res["Value"]["SizeRemoved"]

        # Remove all the files in the directory
        log.debug("Trying to remove %s files." % len(sFilesDict))
        for sFile in sFilesDict:
            # Returns S__OK(Filesize) if it worked
            res = self._removeSingleFile(sFile)

            if res["OK"]:
                filesRemoved += 1
                sizeRemoved += sFilesDict[sFile]["Size"]
            else:
                removedAllFiles = False

        # Check whether all the operations were successful
        allRemoved = removedAllDirs and removedAllFiles

        # Now we try to remove the directory itself
        # We do it only if :
        # If we wanted to remove recursively and everything was deleted
        # We didn't want to remove recursively but we deleted all the files and there are no subfolders

        if (recursive and allRemoved) or (not recursive and removedAllFiles and not subDirsDict):
            try:
                status = self.ctx.rmdir(str(path))
                if status < 0:
                    errStr = "Error occured while removing directory. Status: %s" % status
                    log.debug(errStr)
                    allRemoved = False
            except gfal2.GError as e:
                # How would that be possible...
                if e.code == errno.ENOENT:
                    errStr = "Files does not exist"
                    log.debug(errStr)
                else:
                    errStr = "Failed to remove directory %s" % path
                    log.debug(errStr)
                    allRemoved = False

        resDict = {"AllRemoved": allRemoved, "FilesRemoved": filesRemoved, "SizeRemoved": sizeRemoved}
        return S_OK(resDict)

    def getDirectorySize(self, path):
        """Get the size of the directory on the storage

        .. warning:: it is not recursive

        :param self: self reference
        :param str path: path or list of paths on storage (srm://...)
        :returns: list of successful and failed dictionaries, both indexed by the path

                    * In the failed, the value is the error message
                    * In the successful the values are dictionaries:

                        * Files : amount of files in the dir
                        * Size : summed up size of all files
                        * subDirs : amount of sub dirs

                    * S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.getDirectorySize: Attempting to get size of %s directories" % len(urls))

        failed = {}
        successful = {}

        for url in urls:
            res = self._getSingleDirectorySize(url)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleDirectorySize(self, path):
        """Get the size of the directory on the storage
        CAUTION : the size is not recursive, and does not go into subfolders
        :param self: self reference
        :param path: path (single) on storage (srm://...)
        :return: S_ERROR in case of problem
                  S_OK (Dictionary) Files : amount of files in the directory
                                    Size : summed up size of files
                                    subDirs : amount of sub directories
        """

        self.log.debug("GFAL2_StorageBase._getSingleDirectorySize: Attempting to get the size of directory %s" % path)

        res = self._listSingleDirectory(path)
        if not res["OK"]:
            return res

        directorySize = 0
        directoryFiles = 0
        # itervalues returns a list of values of the dictionnary
        for fileDict in res["Value"]["Files"].values():
            directorySize += fileDict["Size"]
            directoryFiles += 1

        self.log.debug("GFAL2_StorageBase._getSingleDirectorySize: Successfully obtained size of %s." % path)
        subDirectories = len(res["Value"]["SubDirs"])
        return S_OK({"Files": directoryFiles, "Size": directorySize, "SubDirs": subDirectories})

    def getDirectoryMetadata(self, path):
        """Get metadata for the directory(ies) provided

        :param self: self reference
        :param str path: path (or list of paths) on storage (srm://...)
        :returns: Successful dict {path : metadata}
                   Failed dict {path : errStr}
                   S_ERROR in case of argument problems
        """
        res = checkArgumentFormat(path)
        if not res["OK"]:
            return res
        urls = res["Value"]

        self.log.debug("GFAL2_StorageBase.getDirectoryMetadata: Attempting to fetch metadata.")

        failed = {}
        successful = {}

        for url in urls:
            res = self._getSingleDirectoryMetadata(url)

            if not res["OK"]:
                failed[url] = res["Message"]
            else:
                successful[url] = res["Value"]

        return S_OK({"Failed": failed, "Successful": successful})

    def _getSingleDirectoryMetadata(self, path):
        """Fetch the metadata of the provided path

        :param self: self reference
        :param str path: path (only 1) on the storage (srm://...)
        :returns: S_OK( metadataDict ) if we could get the metadata
                  S_ERROR( errStr )if there was a problem getting the metadata or path isn't a directory
        """
        self.log.debug("GFAL2_StorageBase._getSingleDirectoryMetadata: Fetching metadata of directory %s." % path)

        res = self._getSingleMetadata(path)

        if not res["OK"]:
            return res

        metadataDict = res["Value"]

        if not metadataDict["Directory"]:
            errStr = "GFAL2_StorageBase._getSingleDirectoryMetadata: Provided path is not a directory."
            self.log.debug(errStr, path)
            return S_ERROR(errno.ENOTDIR, errStr)

        return S_OK(metadataDict)

    def _getExtendedAttributes(self, path, attributes=None):
        """Get all the available extended attributes of path

        :param self: self reference
        :param str path: path of which we want extended attributes
        :param str list attributes: list of extended attributes we want to receive
        :return: S_OK( attributeDict ) if successful.
                Where the keys of the dict are the attributes and values the respective values
        """

        log = self.log.getSubLogger("GFAL2_StorageBase._getExtendedAttributes")

        log.debug("Checking %s attributes for %s" % (attributes, path))
        attributeDict = {}
        # get all the extended attributes from path
        try:
            if not attributes:
                attributes = self.ctx.listxattr(str(path))

            # get all the respective values of the extended attributes of path
            for attribute in attributes:
                log.debug("Fetching %s" % attribute)
                attributeDict[attribute] = self.ctx.getxattr(str(path), attribute)
            return S_OK(attributeDict)
        # simple error messages, the method that is calling them adds the source of error.
        except gfal2.GError as e:
            errStr = "Something went wrong while checking for extended attributes."
            log.debug(errStr, e.message)
            return S_ERROR(e.code, "%s %s" % (errStr, repr(e)))



#!/usr/bin/env python
"""
Show storage quotas for specified users or for all registered users if nobody is specified

Example:
  $ dirac-admin-user-quota
  ------------------------------
  Username       |     Quota (GB)
  ------------------------------
  atsareg        |           None
  msapunov       |           None
  vhamar         |           None
  ------------------------------
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    # Registering arguments will automatically add their description to the help menu
    Script.registerArgument(["User: list of SEs or comma-separated SEs"], mandatory=False)

    _, users = Script.parseCommandLine()

    from DIRAC import gLogger, gConfig

    if not users:
        res = gConfig.getSections("/Registry/Users")
        if not res["OK"]:
            gLogger.error("Failed to retrieve user list from CS", res["Message"])
            DIRAC.exit(2)
        users = res["Value"]

    gLogger.notice("-" * 30)
    gLogger.notice("%s|%s" % ("Username".ljust(15), "Quota (GB)".rjust(15)))
    gLogger.notice("-" * 30)
    for user in sorted(users):
        quota = gConfig.getValue("/Registry/Users/%s/Quota" % user, 0)
        if not quota:
            quota = gConfig.getValue("/Registry/DefaultStorageQuota")
        gLogger.notice("%s|%s" % (user.ljust(15), str(quota).rjust(15)))
    gLogger.notice("-" * 30)
    DIRAC.exit(0)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient
from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
from DIRAC.DataManagementSystem.Client.DataIntegrityClient import DataIntegrityClient
from DIRAC.DataManagementSystem.Client.DataManager import DataManager

# test 1
AGENT_NAME = "StorageManagement/RequestPreparationAgent"


class RequestPreparationAgent(AgentModule):
    def initialize(self):
        self.fileCatalog = FileCatalog()
        self.dm = DataManager()
        self.stagerClient = StorageManagerClient()
        self.dataIntegrityClient = DataIntegrityClient()
        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")

        return S_OK()

    def execute(self):
        """This is the first logical task to be executed and manages the New->Waiting transition of the Replicas"""
        res = self.__getNewReplicas()
        if not res["OK"]:
            gLogger.fatal(
                "RequestPreparation.prepareNewReplicas: Failed to get replicas from StagerDB.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.info("There were no New replicas found")
            return res
        replicas = res["Value"]["Replicas"]
        replicaIDs = res["Value"]["ReplicaIDs"]
        gLogger.info(
            "RequestPreparation.prepareNewReplicas: Obtained %s New replicas for preparation." % len(replicaIDs)
        )

        # Check if the files exist in the FileCatalog
        res = self.__getExistingFiles(replicas)
        if not res["OK"]:
            return res
        exist = res["Value"]["Exist"]
        terminal = res["Value"]["Missing"]
        failed = res["Value"]["Failed"]
        if not exist:
            gLogger.error("RequestPreparation.prepareNewReplicas: Failed to determine the existence of any file")
            return S_OK()
        terminalReplicaIDs = {}
        for lfn, reason in terminal.items():
            for replicaID in replicas[lfn].values():
                terminalReplicaIDs[replicaID] = reason
            replicas.pop(lfn)
        gLogger.info("RequestPreparation.prepareNewReplicas: %s files exist in the FileCatalog." % len(exist))
        if terminal:
            gLogger.info(
                "RequestPreparation.prepareNewReplicas: %s files do not exist in the FileCatalog." % len(terminal)
            )

        # Obtain the file sizes from the FileCatalog
        res = self.__getFileSize(exist)
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        terminal = res["Value"]["ZeroSize"]
        fileSizes = res["Value"]["FileSizes"]
        if not fileSizes:
            gLogger.error("RequestPreparation.prepareNewReplicas: Failed determine sizes of any files")
            return S_OK()
        for lfn, reason in terminal.items():
            for _se, replicaID in replicas[lfn].items():
                terminalReplicaIDs[replicaID] = reason
            replicas.pop(lfn)
        gLogger.info(
            "RequestPreparation.prepareNewReplicas: Obtained %s file sizes from the FileCatalog." % len(fileSizes)
        )
        if terminal:
            gLogger.info(
                "RequestPreparation.prepareNewReplicas: %s files registered with zero size in the FileCatalog."
                % len(terminal)
            )

        # Obtain the replicas from the FileCatalog
        res = self.__getFileReplicas(list(fileSizes))
        if not res["OK"]:
            return res
        failed.update(res["Value"]["Failed"])
        terminal = res["Value"]["ZeroReplicas"]
        fileReplicas = res["Value"]["Replicas"]
        if not fileReplicas:
            gLogger.error("RequestPreparation.prepareNewReplicas: Failed determine replicas for any files")
            return S_OK()
        for lfn, reason in terminal.items():
            for _se, replicaID in replicas[lfn].items():
                terminalReplicaIDs[replicaID] = reason
            replicas.pop(lfn)
        gLogger.info(
            "RequestPreparation.prepareNewReplicas: Obtained replica information for %s file from the FileCatalog."
            % len(fileReplicas)
        )
        if terminal:
            gLogger.info(
                "RequestPreparation.prepareNewReplicas: %s files registered with zero replicas in the FileCatalog."
                % len(terminal)
            )

        # Check the replicas exist at the requested site
        replicaMetadata = []
        for lfn, requestedSEs in replicas.items():
            lfnReplicas = fileReplicas.get(lfn)

            # This should not happen in principle, but it was seen
            # after a corrupted staging request has entered the DB
            if not lfnReplicas:
                gLogger.error("Missing replicas information", "%s %s" % (lfn, requestedSEs))
                continue

            for requestedSE, replicaID in requestedSEs.items():
                if requestedSE not in lfnReplicas.keys():
                    terminalReplicaIDs[replicaID] = "LFN not registered at requested SE"
                    replicas[lfn].pop(requestedSE)
                else:
                    replicaMetadata.append((replicaID, lfnReplicas[requestedSE], fileSizes[lfn]))

        # Update the states of the files in the database
        if terminalReplicaIDs:
            gLogger.info(
                "RequestPreparation.prepareNewReplicas: %s replicas are terminally failed." % len(terminalReplicaIDs)
            )
            # res = self.stagerClient.updateReplicaFailure( terminalReplicaIDs )
            res = self.stagerClient.updateReplicaFailure(terminalReplicaIDs)
            if not res["OK"]:
                gLogger.error(
                    "RequestPreparation.prepareNewReplicas: Failed to update replica failures.", res["Message"]
                )
        if replicaMetadata:
            gLogger.info(
                "RequestPreparation.prepareNewReplicas: %s replica metadata to be updated." % len(replicaMetadata)
            )
            # Sets the Status='Waiting' of CacheReplicas records that are OK with catalogue checks
            res = self.stagerClient.updateReplicaInformation(replicaMetadata)
            if not res["OK"]:
                gLogger.error(
                    "RequestPreparation.prepareNewReplicas: Failed to update replica metadata.", res["Message"]
                )
        return S_OK()

    def __getNewReplicas(self):
        """This obtains the New replicas from the Replicas table and for each LFN the requested storage element"""
        # First obtain the New replicas from the CacheReplicas table
        res = self.stagerClient.getCacheReplicas({"Status": "New"})
        if not res["OK"]:
            gLogger.error(
                "RequestPreparation.__getNewReplicas: Failed to get replicas with New status.", res["Message"]
            )
            return res
        if not res["Value"]:
            gLogger.debug("RequestPreparation.__getNewReplicas: No New replicas found to process.")
            return S_OK()
        else:
            gLogger.debug(
                "RequestPreparation.__getNewReplicas: Obtained %s New replicas(s) to process." % len(res["Value"])
            )
        replicas = {}
        replicaIDs = {}
        for replicaID, info in res["Value"].items():
            lfn = info["LFN"]
            storageElement = info["SE"]
            replicas.setdefault(lfn, {})[storageElement] = replicaID
            replicaIDs[replicaID] = (lfn, storageElement)
        return S_OK({"Replicas": replicas, "ReplicaIDs": replicaIDs})

    def __getExistingFiles(self, lfns):
        """This checks that the files exist in the FileCatalog."""
        res = self.fileCatalog.exists(list(set(lfns)))
        if not res["OK"]:
            gLogger.error(
                "RequestPreparation.__getExistingFiles: Failed to determine whether files exist.", res["Message"]
            )
            return res
        failed = res["Value"]["Failed"]
        success = res["Value"]["Successful"]
        exist = [lfn for lfn, exists in success.items() if exists]
        missing = list(set(success) - set(exist))
        if missing:
            reason = "LFN not registered in the FC"
            gLogger.warn("RequestPreparation.__getExistingFiles: %s" % reason, "\n".join([""] + missing))
            self.__reportProblematicFiles(missing, "LFN-LFC-DoesntExist")
            missing = dict.fromkeys(missing, reason)
        else:
            missing = {}
        return S_OK({"Exist": exist, "Missing": missing, "Failed": failed})

    def __getFileSize(self, lfns):
        """This obtains the file size from the FileCatalog."""
        fileSizes = {}
        zeroSize = {}
        res = self.fileCatalog.getFileSize(lfns)
        if not res["OK"]:
            gLogger.error("RequestPreparation.__getFileSize: Failed to get sizes for files.", res["Message"])
            return res
        failed = res["Value"]["Failed"]
        for lfn, size in res["Value"]["Successful"].items():
            if size == 0:
                zeroSize[lfn] = "LFN registered with zero size in the FileCatalog"
            else:
                fileSizes[lfn] = size
        if zeroSize:
            for lfn, reason in zeroSize.items():
                gLogger.warn("RequestPreparation.__getFileSize: %s" % reason, lfn)
            self.__reportProblematicFiles(zeroSize.keys(), "LFN-LFC-ZeroSize")
        return S_OK({"FileSizes": fileSizes, "ZeroSize": zeroSize, "Failed": failed})

    def __getFileReplicas(self, lfns):
        """This obtains the replicas from the FileCatalog."""
        replicas = {}
        noReplicas = {}
        res = self.dm.getActiveReplicas(lfns)
        if not res["OK"]:
            gLogger.error("RequestPreparation.__getFileReplicas: Failed to obtain file replicas.", res["Message"])
            return res
        failed = res["Value"]["Failed"]
        for lfn, lfnReplicas in res["Value"]["Successful"].items():
            if len(lfnReplicas) == 0:
                noReplicas[lfn] = "LFN registered with zero replicas in the FileCatalog"
            else:
                replicas[lfn] = lfnReplicas
        if noReplicas:
            for lfn, reason in noReplicas.items():
                gLogger.warn("RequestPreparation.__getFileReplicas: %s" % reason, lfn)
            self.__reportProblematicFiles(list(noReplicas), "LFN-LFC-NoReplicas")
        return S_OK({"Replicas": replicas, "ZeroReplicas": noReplicas, "Failed": failed})

    def __reportProblematicFiles(self, lfns, reason):
        return S_OK()
        res = self.dataIntegrityClient.setFileProblematic(lfns, reason, sourceComponent="RequestPreparationAgent")
        if not res["OK"]:
            gLogger.error(
                "RequestPreparation.__reportProblematicFiles: Failed to report missing files.", res["Message"]
            )
            return res
        if res["Value"]["Successful"]:
            gLogger.info(
                "RequestPreparation.__reportProblematicFiles: Successfully reported %s missing files."
                % len(res["Value"]["Successful"])
            )
        if res["Value"]["Failed"]:
            gLogger.info(
                "RequestPreparation.__reportProblematicFiles: Failed to report %s problematic files."
                % len(res["Value"]["Failed"])
            )
        return res



#!/usr/bin/env python
########################################################################
# File :   dirac-externals-requirements
# Author : Adri/Federico/Andrei
########################################################################
""" If RequiredExternals section is found in releases.cfg of any extension,
    then some python packages to install with pip may be found. This script
    will install the requested modules.

    The command is called from the dirac-install general installation command.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys

# TODO: This should be modernised to use subprocess(32)
try:
    import commands
except ImportError:
    # Python 3's subprocess module contains a compatibility layer
    import subprocess as commands

from diraccfg import CFG

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script
from DIRAC import gLogger, rootPath, S_OK

__RCSID__ = "$Id$"

# Default installation type
instType = "server"


def setInstallType(val):
    global instType
    instType = val
    return S_OK()


def pipInstall(package, switches=""):
    # The right pip should be in the PATH, which is the case after sourcing the DIRAC bashrc
    cmd = "pip install --trusted-host pypi.python.org %s %s" % (switches, package)
    gLogger.notice("Executing %s" % cmd)
    return commands.getstatusoutput(cmd)


@Script()
def main():
    Script.disableCS()

    Script.registerSwitch("t:", "type=", "Installation type. 'server' by default.", setInstallType)
    Script.parseCommandLine(ignoreErrors=True)

    # Collect all the requested python modules to install
    reqDict = {}
    for entry in os.listdir(rootPath):
        if len(entry) < 5 or entry.find("DIRAC") != len(entry) - 5:
            continue
        reqFile = os.path.join(rootPath, entry, "releases.cfg")
        try:
            with open(reqFile, "r") as extfd:
                reqCFG = CFG().loadFromBuffer(extfd.read())
        except Exception:
            gLogger.verbose("%s not found" % reqFile)
            continue
        reqList = reqCFG.getOption("/RequiredExternals/%s" % instType.capitalize(), [])
        if not reqList:
            gLogger.verbose("%s does not have requirements for %s installation" % (entry, instType))
            continue
        for req in reqList:
            reqName = False
            reqCond = ""
            for cond in ("==", ">="):
                iP = cond.find(req)
                if iP > 0:
                    reqName = req[:iP]
                    reqCond = req[iP:]
                    break
            if not reqName:
                reqName = req
            if reqName not in reqDict:
                reqDict[reqName] = (reqCond, entry)
            else:
                gLogger.notice("Skipping %s, it's already requested by %s" % (reqName, reqDict[reqName][1]))

    if not reqDict:
        gLogger.notice("No extra python module requested to be installed")
        sys.exit(0)

    for reqName in reqDict:
        package = "%s%s" % (reqName, reqDict[reqName][0])
        gLogger.notice("Requesting installation of %s" % package)
        status, output = pipInstall(package)
        if status != 0:
            gLogger.error(output)
        else:
            gLogger.notice("Successfully installed %s" % package)


if __name__ == "__main__":
    main()



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, gLogger

from DIRAC.Core.Base.AgentModule import AgentModule
from DIRAC.StorageManagementSystem.Client.StorageManagerClient import StorageManagerClient
from DIRAC.Core.DISET.RPCClient import RPCClient

AGENT_NAME = "StorageManagement/RequestFinalizationAgent"


class RequestFinalizationAgent(AgentModule):
    def initialize(self):

        # This sets the Default Proxy to used as that defined under
        # /Operations/Shifter/DataManager
        # the shifterProxy option in the Configuration can be used to change this default.
        self.am_setOption("shifterProxy", "DataManager")
        self.stagerClient = StorageManagerClient()
        return S_OK()

    def execute(self):
        res = self.clearFailedTasks()
        if not res["OK"]:
            return res
        res = self.callbackStagedTasks()
        if not res["OK"]:
            return res
        res = self.removeUnlinkedReplicas()
        if not res["OK"]:
            return res
        res = self.setOldTasksAsFailed(self.am_getOption("FailIntervalDay", 3))
        return res

    def clearFailedTasks(self):
        """This obtains the tasks which are marked as Failed and remove all the associated records"""
        res = self.stagerClient.getTasksWithStatus("Failed")
        if not res["OK"]:
            gLogger.fatal(
                "RequestFinalization.clearFailedTasks: Failed to get Failed Tasks from StagerDB.", res["Message"]
            )
            return res
        failedTasks = res["Value"]
        gLogger.info(
            "RequestFinalization.clearFailedTasks: Obtained %s tasks in the 'Failed' status." % len(failedTasks)
        )
        for taskID, (_source, callback, sourceTask) in failedTasks.items():
            if callback and sourceTask:
                res = self.__performCallback("Failed", callback, sourceTask)
                if not res["OK"]:
                    failedTasks.pop(taskID)
        if not failedTasks:
            gLogger.info("RequestFinalization.clearFailedTasks: No tasks to remove.")
            return S_OK()
        gLogger.info("RequestFinalization.clearFailedTasks: Removing %s tasks..." % len(failedTasks))
        res = self.stagerClient.removeTasks(list(failedTasks))
        if not res["OK"]:
            gLogger.error("RequestFinalization.clearFailedTasks: Failed to remove tasks.", res["Message"])
            return res
        gLogger.info("RequestFinalization.clearFailedTasks: ...removed.")
        return S_OK()

    def callbackDoneTasks(self):
        """This issues the call back message for the Tasks with a State='Done'"""
        res = self.stagerClient.getTasksWithStatus("Done")
        if not res["OK"]:
            gLogger.fatal(
                "RequestFinalization.callbackDoneTasks: Failed to get Done Tasks from StorageManagementDB.",
                res["Message"],
            )
            return res
        doneTasks = res["Value"]
        gLogger.info("RequestFinalization.callbackDoneTasks: Obtained %s tasks in the 'Done' status." % len(doneTasks))
        for taskID, (_source, callback, sourceTask) in doneTasks.items():
            if callback and sourceTask:
                res = self.__performCallback("Done", callback, sourceTask)
                if not res["OK"]:
                    doneTasks.pop(taskID)
        if not doneTasks:
            gLogger.info("RequestFinalization.callbackDoneTasks: No tasks to update to Done.")
            return S_OK()
        res = self.stagerClient.removeTasks(list(doneTasks))
        if not res["OK"]:
            gLogger.fatal("RequestFinalization.callbackDoneTasks: Failed to remove Done tasks.", res["Message"])
        return res

    def callbackStagedTasks(self):
        """This updates the status of the Tasks to Done then issues the call back message"""
        res = self.stagerClient.getTasksWithStatus("Staged")
        if not res["OK"]:
            gLogger.fatal(
                "RequestFinalization.callbackStagedTasks: Failed to get Staged Tasks from StagerDB.", res["Message"]
            )
            return res
        stagedTasks = res["Value"]
        gLogger.info(
            "RequestFinalization.callbackStagedTasks: Obtained %s tasks in the 'Staged' status." % len(stagedTasks)
        )
        for taskID, (_source, callback, sourceTask) in stagedTasks.items():
            if callback and sourceTask:
                res = self.__performCallback("Done", callback, sourceTask)
                if not res["OK"]:
                    stagedTasks.pop(taskID)
                else:
                    gLogger.info("RequestFinalization.callbackStagedTasks, Task = %s: %s" % (sourceTask, res["Value"]))

        if not stagedTasks:
            gLogger.info("RequestFinalization.callbackStagedTasks: No tasks to update to Done.")
            return S_OK()
        # Daniela: Why is the line below commented out?
        # res = self.stagerClient.setTasksDone(list(stagedTasks))
        res = self.stagerClient.removeTasks(list(stagedTasks))
        if not res["OK"]:
            gLogger.fatal("RequestFinalization.callbackStagedTasks: Failed to remove staged Tasks.", res["Message"])
        return res

    def __performCallback(self, status, callback, sourceTask):
        method, service = callback.split("@")
        gLogger.debug(
            "RequestFinalization.__performCallback: Attempting to perform call back for %s with %s status"
            % (sourceTask, status)
        )
        client = RPCClient(service)
        gLogger.debug("RequestFinalization.__performCallback: Created RPCClient to %s" % service)
        gLogger.debug("RequestFinalization.__performCallback: Attempting to invoke %s service method" % method)
        res = getattr(client, method)(sourceTask, status)
        if not res["OK"]:
            gLogger.error("RequestFinalization.__performCallback: Failed to perform callback", res["Message"])
        else:
            gLogger.info(
                "RequestFinalization.__performCallback: Successfully issued callback to %s for %s with %s status"
                % (callback, sourceTask, status)
            )
        return res

    def removeUnlinkedReplicas(self):
        gLogger.info("RequestFinalization.removeUnlinkedReplicas: Attempting to cleanup unlinked Replicas.")
        res = self.stagerClient.removeUnlinkedReplicas()
        if not res["OK"]:
            gLogger.error(
                "RequestFinalization.removeUnlinkedReplicas: Failed to cleanup unlinked Replicas.", res["Message"]
            )
        else:
            gLogger.info("RequestFinalization.removeUnlinkedReplicas: Successfully removed unlinked Replicas.")
        return res

    def clearReleasedTasks(self):
        # TODO: issue release of the pins associated to this task
        res = self.stagerClient.getTasksWithStatus("Released")
        if not res["OK"]:
            gLogger.fatal(
                "RequestFinalization.clearReleasedTasks: Failed to get Released Tasks from StagerDB.", res["Message"]
            )
            return res
        stagedTasks = res["Value"]
        gLogger.info("RequestFinalization.clearReleasedTasks: Removing %s tasks..." % len(stagedTasks))
        res = self.stagerClient.removeTasks(list(stagedTasks))
        if not res["OK"]:
            gLogger.error("RequestFinalization.clearReleasedTasks: Failed to remove tasks.", res["Message"])
            return res
        gLogger.info("RequestFinalization.clearReleasedTasks: ...removed.")
        return S_OK()

    def setOldTasksAsFailed(self, daysOld):
        gLogger.debug("RequestFinalization.setOldTasksAsFailed: Attempting....")
        res = self.stagerClient.setOldTasksAsFailed(daysOld)
        if not res["OK"]:
            gLogger.error(
                "RequestFinalization.setOldTasksAsFailed: Failed to set old tasks to a Failed state.", res["Message"]
            )
            return res
        return S_OK()



#########################################################################################
# Condor.py
# 10.11.2014
# Author: A.T.
#########################################################################################

""" Condor.py is a DIRAC independent class representing Condor batch system.
    Condor objects are used as backend batch system representation for
    LocalComputingElement and SSHComputingElement classes
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import re
import tempfile
import subprocess
import shlex
import os

__RCSID__ = "$Id$"


def parseCondorStatus(lines, jobID):
    """parse the condor_q or condor_history output for the job status

    :param lines: list of lines from the output of the condor commands, each line is a pair of jobID and statusID
    :type lines: python:list
    :param str jobID: jobID of condor job, e.g.: 123.53
    :returns: Status as known by DIRAC
    """
    jobID = str(jobID)
    for line in lines:
        l = line.strip().split()
        try:
            status = int(l[1])
        except (ValueError, IndexError):
            continue
        if l[0] == jobID:
            return {1: "Waiting", 2: "Running", 3: "Aborted", 4: "Done", 5: "HELD"}.get(status, "Unknown")
    return "Unknown"


def treatCondorHistory(condorHistCall, qList):
    """concatenate clusterID and processID to get the same output as condor_q
    until we can expect condor version 8.5.3 everywhere

    :param str condorHistCall: condor_history command to run
    :param qList: list of jobID and status from condor_q output, will be modified in this function
    :type qList: python:list
    :returns: None
    """
    sp = subprocess.Popen(
        shlex.split(condorHistCall),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )
    output, _ = sp.communicate()
    status = sp.returncode

    # Join the ClusterId and the ProcId and add to existing list of statuses
    if status == 0:
        for line in output.split("\n"):
            values = line.strip().split()
            if len(values) == 3:
                qList.append("%s.%s %s" % tuple(values))


class Condor(object):
    def submitJob(self, **kwargs):
        """Submit nJobs to the Condor batch system"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["Executable", "OutputDir", "SubmitOptions"]

        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        nJobs = kwargs.get("NJobs")
        if not nJobs:
            nJobs = 1
        numberOfProcessors = kwargs.get("NumberOfProcessors")
        wholeNode = kwargs.get("WholeNode")
        outputDir = kwargs["OutputDir"]
        executable = kwargs["Executable"]
        submitOptions = kwargs["SubmitOptions"]
        preamble = kwargs.get("Preamble")

        if wholeNode:
            requirements = (
                '+RequiresWholeMachine=True\n Requirements = ( CAN_RUN_WHOLE_MACHINE ) && ( OpSys == "LINUX" )'
            )
        else:
            requirements = 'Requirements = OpSys == "LINUX"'

        jdlFile = tempfile.NamedTemporaryFile(dir=outputDir, suffix=".jdl")
        jdlFile.write(
            """
    Executable = %s
    Universe = vanilla
    %s
    Initialdir = %s
    Output = $(Cluster).$(Process).out
    Error = $(Cluster).$(Process).err
    Log = test.log
    Environment = CONDOR_JOBID=$(Cluster).$(Process)
    Getenv = False

    request_cpus = %s

    Queue %s

    """
            % (executable, requirements, outputDir, numberOfProcessors, nJobs)
        )

        jdlFile.flush()

        cmd = "%s; " % preamble if preamble else ""
        cmd += "condor_submit %s %s" % (submitOptions, jdlFile.name)
        sp = subprocess.Popen(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        jdlFile.close()

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        submittedJobs = 0
        cluster = ""
        lines = output.split("\n")
        for line in lines:
            if "cluster" in line:
                result = re.match(r"(\d+) job.*cluster (\d+)\.", line)
                if result:
                    submittedJobs, cluster = result.groups()
                    try:
                        submittedJobs = int(submittedJobs)
                    except BaseException:
                        submittedJobs = 0

        if submittedJobs > 0 and cluster:
            resultDict["Status"] = 0
            resultDict["Jobs"] = []
            for i in range(submittedJobs):
                resultDict["Jobs"].append(".".join([cluster, str(i)]))
            # Executable is transferred afterward
            # Inform the caller that Condor cannot delete it before the end of the execution
            resultDict["ExecutableToKeep"] = executable
        else:
            resultDict["Status"] = status
            resultDict["Message"] = error
        return resultDict

    def killJob(self, **kwargs):
        """Kill jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        successful = []
        failed = []
        errors = ""
        for job in jobIDList:
            sp = subprocess.Popen(
                shlex.split("condor_rm %s" % job),
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
            )
            output, error = sp.communicate()
            status = sp.returncode
            if status != 0:
                failed.append(job)
                errors += error
            else:
                successful.append(job)

        resultDict["Status"] = 0
        if failed:
            resultDict["Status"] = 1
            resultDict["Message"] = errors
        resultDict["Successful"] = successful
        resultDict["Failed"] = failed
        return resultDict

    def getJobStatus(self, **kwargs):
        """Get status of the jobs in the given list"""

        resultDict = {}

        MANDATORY_PARAMETERS = ["JobIDList"]
        for argument in MANDATORY_PARAMETERS:
            if argument not in kwargs:
                resultDict["Status"] = -1
                resultDict["Message"] = "No %s" % argument
                return resultDict

        jobIDList = kwargs["JobIDList"]
        if not jobIDList:
            resultDict["Status"] = -1
            resultDict["Message"] = "Empty job list"
            return resultDict

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        cmd = "condor_q -submitter %s -af:j JobStatus" % user
        sp = subprocess.Popen(
            shlex.split(cmd),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        qList = output.strip().split("\n")

        # FIXME: condor_history does only support j for autoformat from 8.5.3,
        # format adds whitespace for each field This will return a list of 1245 75 3
        # needs to cocatenate the first two with a dot
        condorHistCall = "condor_history -af ClusterId ProcId JobStatus -submitter %s" % user
        treatCondorHistory(condorHistCall, qList)

        statusDict = {}
        if len(qList):
            for job in jobIDList:
                job = str(job)
                statusDict[job] = parseCondorStatus(qList, job)
                if statusDict[job] == "HELD":
                    statusDict[job] = "Unknown"

        # Final output
        status = 0
        resultDict["Status"] = 0
        resultDict["Jobs"] = statusDict
        return resultDict

    def getCEStatus(self, **kwargs):
        """Get the overall status of the CE"""
        resultDict = {}

        user = kwargs.get("User")
        if not user:
            user = os.environ.get("USER")
        if not user:
            resultDict["Status"] = -1
            resultDict["Message"] = "No user name"
            return resultDict

        waitingJobs = 0
        runningJobs = 0

        sp = subprocess.Popen(
            shlex.split("condor_q -submitter %s" % user),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        output, error = sp.communicate()
        status = sp.returncode

        if status != 0:
            if "no record" in output:
                resultDict["Status"] = 0
                resultDict["Waiting"] = waitingJobs
                resultDict["Running"] = runningJobs
                return resultDict
            resultDict["Status"] = status
            resultDict["Message"] = error
            return resultDict

        if "no record" in output:
            resultDict["Status"] = 0
            resultDict["Waiting"] = waitingJobs
            resultDict["Running"] = runningJobs
            return resultDict

        if output:
            lines = output.split("\n")
            for line in lines:
                if not line.strip():
                    continue
                if " I " in line:
                    waitingJobs += 1
                elif " R " in line:
                    runningJobs += 1

        # Final output
        resultDict["Status"] = 0
        resultDict["Waiting"] = waitingJobs
        resultDict["Running"] = runningJobs
        return resultDict



""" Test for WMS clients
"""
# pylint: disable=protected-access, missing-docstring, invalid-name, line-too-long

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import unittest
import importlib

from six import BytesIO

from mock import MagicMock

# sut
from DIRAC.WorkloadManagementSystem.Client.Matcher import Matcher
from DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient import SandboxStoreClient


class ClientsTestCase(unittest.TestCase):
    """Base class for the clients test cases"""

    def setUp(self):

        from DIRAC import gLogger

        gLogger.setLevel("DEBUG")

        self.pilotAgentsDBMock = MagicMock()
        self.jobDBMock = MagicMock()
        self.tqDBMock = MagicMock()
        self.jlDBMock = MagicMock()
        self.opsHelperMock = MagicMock()
        self.matcher = Matcher(
            pilotAgentsDB=self.pilotAgentsDBMock,
            jobDB=self.jobDBMock,
            tqDB=self.tqDBMock,
            jlDB=self.jlDBMock,
            opsHelper=self.opsHelperMock,
        )

    def tearDown(self):
        try:
            os.remove("1.txt")
            os.remove("InputData_*")
        except OSError:
            pass


class MatcherTestCase(ClientsTestCase):
    def test__processResourceDescription(self):

        resourceDescription = {
            "Architecture": "x86_64-slc6",
            "CEQueue": "jenkins-queue_not_important",
            "CPUNormalizationFactor": "9.5",
            "CPUTime": 1080000,
            "CPUTimeLeft": 5000,
            "DIRACVersion": "v8r0p1",
            "FileCatalog": "LcgFileCatalogCombined",
            "GridCE": "jenkins.cern.ch",
            "LocalSE": ["CERN-SWTEST"],
            "MaxTotalJobs": 100,
            "MaxWaitingJobs": 10,
            "OutputURL": "gsiftp://localhost",
            "PilotBenchmark": 9.5,
            "PilotReference": "somePilotReference",
            "Platform": "x86_64-slc6",
            "ReleaseVersion": "v8r0p1",
            "Setup": "LHCb-Certification",
            "Site": "DIRAC.Jenkins.ch",
            "WaitingToRunningRatio": 0.05,
        }

        res = self.matcher._processResourceDescription(resourceDescription)
        resExpected = {
            "Setup": "LHCb-Certification",
            "ReleaseVersion": "v8r0p1",
            "CPUTime": 1080000,
            "DIRACVersion": "v8r0p1",
            "PilotReference": "somePilotReference",
            "PilotBenchmark": 9.5,
            "Platform": "x86_64-slc6",
            "Site": "DIRAC.Jenkins.ch",
            "GridCE": "jenkins.cern.ch",
        }

        self.assertEqual(res, resExpected)


#############################################################################


class SandboxStoreTestCaseSuccess(ClientsTestCase):
    def test_uploadFilesAsSandbox(self):

        ourSSC = importlib.import_module("DIRAC.WorkloadManagementSystem.Client.SandboxStoreClient")
        ourSSC.TransferClient = MagicMock()
        ssc = SandboxStoreClient()
        fileList = [BytesIO(b"try")]
        res = ssc.uploadFilesAsSandbox(fileList)
        print(res)


#############################################################################
# Test Suite run
#############################################################################

if __name__ == "__main__":
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(ClientsTestCase)
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(MatcherTestCase))
    suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(SandboxStoreTestCaseSuccess))
    testResult = unittest.TextTestRunner(verbosity=2).run(suite)

# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" Synchronizer

  Module that keeps the database synchronized with the CS
  Module that updates the RSS database ( ResourceStatusDB ) with the information
  in the Resources section. If there are additions in the CS, those are incorporated
  to the DB. If there are deletions, entries in RSS tables for those elements are
  deleted ( except the Logs table ).

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import gLogger, S_OK
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.Core.Utilities.SiteSEMapping import getStorageElementsHosts
from DIRAC.ConfigurationSystem.Client.Helpers.Resources import getSites, getFTS3Servers, getCESiteMapping
from DIRAC.ConfigurationSystem.Client.PathFinder import getServiceURL
from DIRAC.DataManagementSystem.Utilities.DMSHelpers import DMSHelpers
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Utilities import CSHelpers
from DIRAC.ResourceStatusSystem.Utilities.RssConfiguration import RssConfiguration
from DIRAC.ResourceStatusSystem.Utilities import Utils

ResourceManagementClient = getattr(
    Utils.voimport("DIRAC.ResourceStatusSystem.Client.ResourceManagementClient"), "ResourceManagementClient"
)


class Synchronizer(object):
    """
    Every time there is a successful write on the CS, Synchronizer().sync() is
    executed. It updates the database with the values on the CS.

    """

    def __init__(self, rStatus=None, rManagement=None, defaultStatus="Unknown"):

        # Warm up local CS
        CSHelpers.warmUp()

        if rStatus is None:
            self.rStatus = ResourceStatusClient()
        if rManagement is None:
            self.rManagement = ResourceManagementClient()
        self.defaultStatus = defaultStatus

        self.rssConfig = RssConfiguration()

        # this just sets the main owner, "rs_svc" just mean "RSS service"
        self.tokenOwner = "rs_svc"

        # if we are running this script as a user (from a CLI),
        # the username found the proxy will be used as tokenOwner
        result = getProxyInfo()
        if result["OK"]:
            self.tokenOwner = result["Value"]["username"]

    def sync(self, _eventName, _params):
        """
        Main synchronizer method. It synchronizes the three types of elements: Sites,
        Resources and Nodes. Each _syncX method returns a dictionary with the additions
        and deletions.

        examples:
          >>> s.sync( None, None )
              S_OK()

        :Parameters:
          **_eventName** - any
            this parameter is ignored, but needed by caller function.
          **_params** - any
            this parameter is ignored, but needed by caller function.

        :return: S_OK
        """

        syncSites = self._syncSites()
        if not syncSites["OK"]:
            gLogger.error(syncSites["Message"])

        syncResources = self._syncResources()
        if not syncResources["OK"]:
            gLogger.error(syncResources["Message"])

        syncNodes = self._syncNodes()
        if not syncNodes["OK"]:
            gLogger.error(syncNodes["Message"])

        return S_OK()

    def _syncSites(self):
        """
        Sync sites: compares CS with DB and does the necessary modifications.
        """

        gLogger.info("-- Synchronizing sites --")

        # sites in CS
        res = getSites()
        if not res["OK"]:
            return res
        sitesCS = res["Value"]

        gLogger.verbose("%s sites found in CS" % len(sitesCS))

        # sites in RSS
        result = self.rStatus.selectStatusElement("Site", "Status", meta={"columns": ["Name"]})
        if not result["OK"]:
            return result
        sitesDB = [siteDB[0] for siteDB in result["Value"]]

        # Sites that are in DB but not (anymore) in CS
        toBeDeleted = list(set(sitesDB).difference(set(sitesCS)))
        gLogger.verbose("%s sites to be deleted" % len(toBeDeleted))

        # Delete sites
        for siteName in toBeDeleted:
            deleteQuery = self.rStatus._extermineStatusElement("Site", siteName)
            gLogger.verbose("Deleting site %s" % siteName)
            if not deleteQuery["OK"]:
                return deleteQuery

        # Sites that are in CS but not (anymore) in DB
        toBeAdded = list(set(sitesCS).difference(set(sitesDB)))
        gLogger.verbose("%s site entries to be added" % len(toBeAdded))

        for site in toBeAdded:
            query = self.rStatus.addIfNotThereStatusElement(
                "Site",
                "Status",
                name=site,
                statusType="all",
                status=self.defaultStatus,
                elementType="Site",
                tokenOwner=self.tokenOwner,
                reason="Synchronized",
            )
            if not query["OK"]:
                return query

        return S_OK()

    def _syncResources(self):
        """
        Sync resources: compares CS with DB and does the necessary modifications.
        ( StorageElements, FTS, FileCatalogs and ComputingElements )
        """

        gLogger.info("-- Synchronizing Resources --")

        gLogger.verbose("-> StorageElements")
        ses = self.__syncStorageElements()
        if not ses["OK"]:
            gLogger.error(ses["Message"])

        gLogger.verbose("-> FTS")
        fts = self.__syncFTS()
        if not fts["OK"]:
            gLogger.error(fts["Message"])

        gLogger.verbose("-> FileCatalogs")
        fileCatalogs = self.__syncFileCatalogs()
        if not fileCatalogs["OK"]:
            gLogger.error(fileCatalogs["Message"])

        gLogger.verbose("-> ComputingElements")
        computingElements = self.__syncComputingElements()
        if not computingElements["OK"]:
            gLogger.error(computingElements["Message"])

        gLogger.verbose("-> removing resources that no longer exist in the CS")
        removingResources = self.__removeNonExistingResourcesFromRM()
        if not removingResources["OK"]:
            gLogger.error(removingResources["Message"])

        return S_OK()

    def _syncNodes(self):
        """
        Sync resources: compares CS with DB and does the necessary modifications.
        ( Queues )
        """
        gLogger.info("-- Synchronizing Nodes --")

        gLogger.verbose("-> Queues")
        queues = self.__syncQueues()
        if not queues["OK"]:
            gLogger.error(queues["Message"])

        return S_OK()

    def __removeNonExistingResourcesFromRM(self):
        """
        Remove resources from DowntimeCache table that no longer exist in the CS.
        """

        if not getServiceURL("ResourceStatus/ResourceManagement"):
            gLogger.verbose("ResourceManagement is not installed, skipping removal of non existing resources...")
            return S_OK()

        sesHosts = getStorageElementsHosts()
        if not sesHosts["OK"]:
            return sesHosts
        sesHosts = sesHosts["Value"]

        resources = sesHosts

        ftsServer = getFTS3Servers(hostOnly=True)
        if ftsServer["OK"]:
            resources.extend(ftsServer["Value"])

        res = getCESiteMapping()
        if res["OK"]:
            resources.extend(list(res["Value"]))

        downtimes = self.rManagement.selectDowntimeCache()

        if not downtimes["OK"]:
            return downtimes

        # Remove hosts that no longer exist in the CS
        for host in downtimes["Value"]:
            gLogger.verbose("Checking if %s is still in the CS" % host[0])
            if host[0] not in resources:
                gLogger.verbose("%s is no longer in CS, removing entry..." % host[0])
                result = self.rManagement.deleteDowntimeCache(name=host[0])

                if not result["OK"]:
                    return result

        return S_OK()

    def __syncComputingElements(self):
        """
        Sync ComputingElements: compares CS with DB and does the necessary modifications.
        """

        res = getCESiteMapping()
        if not res["OK"]:
            return res
        cesCS = list(res["Value"])

        gLogger.verbose("%s Computing elements found in CS" % len(cesCS))

        cesDB = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="ComputingElement", meta={"columns": ["Name"]}
        )
        if not cesDB["OK"]:
            return cesDB
        cesDB = [ceDB[0] for ceDB in cesDB["Value"]]

        # ComputingElements that are in DB but not in CS
        toBeDeleted = list(set(cesDB).difference(set(cesCS)))
        gLogger.verbose("%s Computing elements to be deleted" % len(toBeDeleted))

        # Delete storage elements
        for ceName in toBeDeleted:

            deleteQuery = self.rStatus._extermineStatusElement("Resource", ceName)

            gLogger.verbose("... %s" % ceName)
            if not deleteQuery["OK"]:
                return deleteQuery

        # statusTypes = RssConfiguration.getValidStatusTypes()[ 'Resource' ]
        statusTypes = self.rssConfig.getConfigStatusType("ComputingElement")

        result = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="ComputingElement", meta={"columns": ["Name", "StatusType"]}
        )
        if not result["OK"]:
            return result
        cesTuple = [(x[0], x[1]) for x in result["Value"]]

        # For each ( se, statusType ) tuple not present in the DB, add it.
        cesStatusTuples = [(se, statusType) for se in cesCS for statusType in statusTypes]
        toBeAdded = list(set(cesStatusTuples).difference(set(cesTuple)))

        gLogger.debug("%s Computing elements entries to be added" % len(toBeAdded))

        for ceTuple in toBeAdded:

            _name = ceTuple[0]
            _statusType = ceTuple[1]
            _status = self.defaultStatus
            _reason = "Synchronized"
            _elementType = "ComputingElement"

            query = self.rStatus.addIfNotThereStatusElement(
                "Resource",
                "Status",
                name=_name,
                statusType=_statusType,
                status=_status,
                elementType=_elementType,
                tokenOwner=self.tokenOwner,
                reason=_reason,
            )
            if not query["OK"]:
                return query

        return S_OK()

    def __syncFileCatalogs(self):
        """
        Sync FileCatalogs: compares CS with DB and does the necessary modifications.
        """

        catalogsCS = CSHelpers.getFileCatalogs()
        if not catalogsCS["OK"]:
            return catalogsCS
        catalogsCS = catalogsCS["Value"]

        gLogger.verbose("%s File catalogs found in CS" % len(catalogsCS))

        catalogsDB = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="Catalog", meta={"columns": ["Name"]}
        )
        if not catalogsDB["OK"]:
            return catalogsDB
        catalogsDB = [catalogDB[0] for catalogDB in catalogsDB["Value"]]

        # StorageElements that are in DB but not in CS
        toBeDeleted = list(set(catalogsDB).difference(set(catalogsCS)))
        gLogger.verbose("%s File catalogs to be deleted" % len(toBeDeleted))

        # Delete storage elements
        for catalogName in toBeDeleted:

            deleteQuery = self.rStatus._extermineStatusElement("Resource", catalogName)

            gLogger.verbose("... %s" % catalogName)
            if not deleteQuery["OK"]:
                return deleteQuery

        # statusTypes = RssConfiguration.getValidStatusTypes()[ 'Resource' ]
        statusTypes = self.rssConfig.getConfigStatusType("Catalog")

        result = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="Catalog", meta={"columns": ["Name", "StatusType"]}
        )
        if not result["OK"]:
            return result
        sesTuple = [(x[0], x[1]) for x in result["Value"]]

        # For each ( se, statusType ) tuple not present in the DB, add it.
        catalogsStatusTuples = [(se, statusType) for se in catalogsCS for statusType in statusTypes]
        toBeAdded = list(set(catalogsStatusTuples).difference(set(sesTuple)))

        gLogger.verbose("%s File catalogs entries to be added" % len(toBeAdded))

        for catalogTuple in toBeAdded:

            _name = catalogTuple[0]
            _statusType = catalogTuple[1]
            _status = self.defaultStatus
            _reason = "Synchronized"
            _elementType = "Catalog"

            query = self.rStatus.addIfNotThereStatusElement(
                "Resource",
                "Status",
                name=_name,
                statusType=_statusType,
                status=_status,
                elementType=_elementType,
                tokenOwner=self.tokenOwner,
                reason=_reason,
            )
            if not query["OK"]:
                return query

        return S_OK()

    def __syncFTS(self):
        """
        Sync FTS: compares CS with DB and does the necessary modifications.
        """

        ftsCS = CSHelpers.getFTS()
        if not ftsCS["OK"]:
            return ftsCS
        ftsCS = ftsCS["Value"]

        gLogger.verbose("%s FTS endpoints found in CS" % len(ftsCS))

        ftsDB = self.rStatus.selectStatusElement("Resource", "Status", elementType="FTS", meta={"columns": ["Name"]})
        if not ftsDB["OK"]:
            return ftsDB
        ftsDB = [fts[0] for fts in ftsDB["Value"]]

        # StorageElements that are in DB but not in CS
        toBeDeleted = list(set(ftsDB).difference(set(ftsCS)))
        gLogger.verbose("%s FTS endpoints to be deleted" % len(toBeDeleted))

        # Delete storage elements
        for ftsName in toBeDeleted:

            deleteQuery = self.rStatus._extermineStatusElement("Resource", ftsName)

            gLogger.verbose("... %s" % ftsName)
            if not deleteQuery["OK"]:
                return deleteQuery

        statusTypes = self.rssConfig.getConfigStatusType("FTS")
        # statusTypes = RssConfiguration.getValidStatusTypes()[ 'Resource' ]

        result = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="FTS", meta={"columns": ["Name", "StatusType"]}
        )
        if not result["OK"]:
            return result
        sesTuple = [(x[0], x[1]) for x in result["Value"]]

        # For each ( se, statusType ) tuple not present in the DB, add it.
        ftsStatusTuples = [(se, statusType) for se in ftsCS for statusType in statusTypes]
        toBeAdded = list(set(ftsStatusTuples).difference(set(sesTuple)))

        gLogger.verbose("%s FTS endpoints entries to be added" % len(toBeAdded))

        for ftsTuple in toBeAdded:

            _name = ftsTuple[0]
            _statusType = ftsTuple[1]
            _status = self.defaultStatus
            _reason = "Synchronized"
            _elementType = "FTS"

            query = self.rStatus.addIfNotThereStatusElement(
                "Resource",
                "Status",
                name=_name,
                statusType=_statusType,
                status=_status,
                elementType=_elementType,
                tokenOwner=self.tokenOwner,
                reason=_reason,
            )
            if not query["OK"]:
                return query

        return S_OK()

    def __syncStorageElements(self):
        """
        Sync StorageElements: compares CS with DB and does the necessary modifications.
        """

        sesCS = DMSHelpers().getStorageElements()

        gLogger.verbose("%s storage elements found in CS" % len(sesCS))

        sesDB = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="StorageElement", meta={"columns": ["Name"]}
        )
        if not sesDB["OK"]:
            return sesDB
        sesDB = [seDB[0] for seDB in sesDB["Value"]]

        # StorageElements that are in DB but not in CS
        toBeDeleted = list(set(sesDB).difference(set(sesCS)))
        gLogger.verbose("%s storage elements to be deleted" % len(toBeDeleted))

        # Delete storage elements
        for sesName in toBeDeleted:

            deleteQuery = self.rStatus._extermineStatusElement("Resource", sesName)

            gLogger.verbose("... %s" % sesName)
            if not deleteQuery["OK"]:
                return deleteQuery

        statusTypes = self.rssConfig.getConfigStatusType("StorageElement")
        # statusTypes = RssConfiguration.getValidStatusTypes()[ 'Resource' ]

        result = self.rStatus.selectStatusElement(
            "Resource", "Status", elementType="StorageElement", meta={"columns": ["Name", "StatusType"]}
        )
        if not result["OK"]:
            return result
        sesTuple = [(x[0], x[1]) for x in result["Value"]]

        # For each ( se, statusType ) tuple not present in the DB, add it.
        sesStatusTuples = [(se, statusType) for se in sesCS for statusType in statusTypes]
        toBeAdded = list(set(sesStatusTuples).difference(set(sesTuple)))

        gLogger.verbose("%s storage element entries to be added" % len(toBeAdded))

        for seTuple in toBeAdded:

            _name = seTuple[0]
            _statusType = seTuple[1]
            _status = self.defaultStatus
            _reason = "Synchronized"
            _elementType = "StorageElement"

            query = self.rStatus.addIfNotThereStatusElement(
                "Resource",
                "Status",
                name=_name,
                statusType=_statusType,
                status=_status,
                elementType=_elementType,
                tokenOwner=self.tokenOwner,
                reason=_reason,
            )
            if not query["OK"]:
                return query

        return S_OK()

    def __syncQueues(self):
        """
        Sync Queues: compares CS with DB and does the necessary modifications.
        """

        queuesCS = CSHelpers.getQueuesRSS()
        if not queuesCS["OK"]:
            return queuesCS
        queuesCS = queuesCS["Value"]

        gLogger.verbose("%s Queues found in CS" % len(queuesCS))

        queuesDB = self.rStatus.selectStatusElement("Node", "Status", elementType="Queue", meta={"columns": ["Name"]})
        if not queuesDB["OK"]:
            return queuesDB
        queuesDB = [queueDB[0] for queueDB in queuesDB["Value"]]

        # ComputingElements that are in DB but not in CS
        toBeDeleted = list(set(queuesDB).difference(set(queuesCS)))
        gLogger.verbose("%s Queues to be deleted" % len(toBeDeleted))

        # Delete storage elements
        for queueName in toBeDeleted:

            deleteQuery = self.rStatus._extermineStatusElement("Node", queueName)

            gLogger.verbose("... %s" % queueName)
            if not deleteQuery["OK"]:
                return deleteQuery

        statusTypes = self.rssConfig.getConfigStatusType("Queue")
        # statusTypes = RssConfiguration.getValidStatusTypes()[ 'Node' ]

        result = self.rStatus.selectStatusElement(
            "Node", "Status", elementType="Queue", meta={"columns": ["Name", "StatusType"]}
        )
        if not result["OK"]:
            return result
        queueTuple = [(x[0], x[1]) for x in result["Value"]]

        # For each ( se, statusType ) tuple not present in the DB, add it.
        queueStatusTuples = [(se, statusType) for se in queuesCS for statusType in statusTypes]
        toBeAdded = list(set(queueStatusTuples).difference(set(queueTuple)))

        gLogger.verbose("%s Queue entries to be added" % len(toBeAdded))

        for queueTuple in toBeAdded:

            _name = queueTuple[0]
            _statusType = queueTuple[1]
            _status = self.defaultStatus
            _reason = "Synchronized"
            _elementType = "Queue"

            query = self.rStatus.addIfNotThereStatusElement(
                "Node",
                "Status",
                name=_name,
                statusType=_statusType,
                status=_status,
                elementType=_elementType,
                tokenOwner=self.tokenOwner,
                reason=_reason,
            )
            if not query["OK"]:
                return query

        return S_OK()



#!/usr/bin/env python
########################################################################
# File :    dirac-wms-select-jobs
# Author :  Stuart Paterson
########################################################################
"""
Select DIRAC jobs matching the given conditions
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import DIRAC
from DIRAC import gLogger
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    maxJobs = 100
    Script.registerSwitch("", "Status=", "Primary status")
    Script.registerSwitch("", "MinorStatus=", "Secondary status")
    Script.registerSwitch("", "ApplicationStatus=", "Application status")
    Script.registerSwitch("", "Site=", "Execution site")
    Script.registerSwitch("", "Owner=", "Owner (DIRAC nickname)")
    Script.registerSwitch("", "JobGroup=", "Select jobs for specified job group")
    Script.registerSwitch("", "Date=", "Date in YYYY-MM-DD format, if not specified default is today")
    Script.registerSwitch("", "Maximum=", "Maximum number of jobs shown (default %d, 0 means all)" % maxJobs)
    switches, args = Script.parseCommandLine(ignoreErrors=True)

    # Default values
    status = None
    minorStatus = None
    appStatus = None
    site = None
    owner = None
    jobGroups = []
    date = None

    if args:
        Script.showHelp()

    exitCode = 0

    for switch in switches:
        if switch[0].lower() == "status":
            status = switch[1]
        elif switch[0].lower() == "minorstatus":
            minorStatus = switch[1]
        elif switch[0].lower() == "applicationstatus":
            appStatus = switch[1]
        elif switch[0].lower() == "site":
            site = switch[1]
        elif switch[0].lower() == "owner":
            owner = switch[1]
        elif switch[0].lower() == "jobgroup":
            for jg in switch[1].split(","):
                if jg.isdigit():
                    jobGroups.append("%08d" % int(jg))
                else:
                    jobGroups.append(jg)
        elif switch[0].lower() == "date":
            date = switch[1]
        elif switch[0] == "Maximum":
            try:
                maxJobs = int(switch[1])
            except TypeError:
                gLogger.fatal("Invalid max number of jobs", switch[1])
                DIRAC.exit(1)

    selDate = date
    if not date:
        selDate = "Today"
    conditions = {
        "Status": status,
        "MinorStatus": minorStatus,
        "ApplicationStatus": appStatus,
        "Owner": owner,
        "JobGroup": ",".join(str(jg) for jg in jobGroups),
        "Date": selDate,
    }

    from DIRAC.Interfaces.API.Dirac import Dirac

    dirac = Dirac()
    jobs = []

    if jobGroups:
        for jobGroup in jobGroups:
            res = dirac.selectJobs(
                status=status,
                minorStatus=minorStatus,
                applicationStatus=appStatus,
                site=site,
                owner=owner,
                jobGroup=jobGroup,
                date=date,
                printErrors=False,
            )
            if res["OK"]:
                jobs.extend(res["Value"])
            else:
                gLogger.error("Can't select jobs: ", res["Message"])
    else:
        res = dirac.selectJobs(
            status=status,
            minorStatus=minorStatus,
            applicationStatus=appStatus,
            site=site,
            owner=owner,
            date=date,
            printErrors=False,
        )
        if res["OK"]:
            jobs.extend(res["Value"])
        else:
            gLogger.error("Can't select jobs: ", res["Message"])

    conds = ["%s = %s" % (n, v) for n, v in conditions.items() if v]
    if maxJobs and len(jobs) > maxJobs:
        jobs = jobs[:maxJobs]
        constrained = " (first %d shown) " % maxJobs
    else:
        constrained = " "

    if jobs:
        gLogger.notice(
            "==> Selected %s jobs%swith conditions: %s\n%s" % (len(jobs), constrained, ", ".join(conds), ",".join(jobs))
        )
    else:
        gLogger.notice("No jobs were selected with conditions:", ", ".join(conds))

    DIRAC.exit(exitCode)


if __name__ == "__main__":
    main()



""" Configurations module

  Configuration to use policies.

  Follows the schema::

    <PolicyNameInCS> : {
               'description' : <some human readable description>,
               'module'      : <policy module name>,
               'command'     : ( <command module name >, < command class name > ),
               'args'        : { arguments for the command } or None
                       }

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id: $"


POLICIESMETA = {  # DownTime POLICIES
    "DTOngoing": {
        "description": "Ongoing and scheduled down-times",
        "module": "DowntimePolicy",
        "command": ("DowntimeCommand", "DowntimeCommand"),
        "args": {"hours": 0, "onlyCache": True},
    },
    "DTScheduled1": {
        "description": "Ongoing and scheduled down-times",
        "module": "DowntimePolicy",
        "command": ("DowntimeCommand", "DowntimeCommand"),
        "args": {"hours": 1, "onlyCache": True},
    },
    "DTScheduled3": {
        "description": "Ongoing and scheduled down-times",
        "module": "DowntimePolicy",
        "command": ("DowntimeCommand", "DowntimeCommand"),
        "args": {"hours": 3, "onlyCache": True},
    },
    "DTScheduled": {
        "description": "Scheduled down-times, starting in <hours>",
        "module": "DowntimePolicy",
        "command": ("DowntimeCommand", "DowntimeCommand"),
        "args": {"hours": 12, "onlyCache": True},
    },
    # Free Disk Space in Terabytes
    "FreeDiskSpaceTB": {
        "description": "Free disk space, in TB",
        "module": "FreeDiskSpacePolicy",
        "command": ("FreeDiskSpaceCommand", "FreeDiskSpaceCommand"),
        "args": {"unit": "TB", "onlyCache": True},
    },
    # Free Disk Space in Gigabytes
    "FreeDiskSpaceGB": {
        "description": "Free disk space, in GB",
        "module": "FreeDiskSpacePolicy",
        "command": ("FreeDiskSpaceCommand", "FreeDiskSpaceCommand"),
        "args": {"unit": "GB", "onlyCache": True},
    },
    # Free Disk Space in Megabytes
    "FreeDiskSpaceMB": {
        "description": "Free disk space, in MB",
        "module": "FreeDiskSpacePolicy",
        "command": ("FreeDiskSpaceCommand", "FreeDiskSpaceCommand"),
        "args": {"unit": "MB", "onlyCache": True},
    },
    # GGUS tickets open
    "GGUSTickets": {
        "description": "Open GGUS tickets",
        "module": "GGUSTicketsPolicy",
        "command": ("GGUSTicketsCommand", "GGUSTicketsCommand"),
        "args": {"onlyCache": False},
    },
    # Job POLICIES
    "JobDoneRatio": {
        "description": "done / ( completed + done ) jobs ( 30 min )",
        "module": "JobDoneRatioPolicy",
        "command": ("JobCommand", "JobCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    "JobEfficiency": {
        "description": "( completed + done ) / ( completed + done + failed ) jobs ( 30 min )",
        "module": "JobEfficiencyPolicy",
        "command": ("JobCommand", "JobCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    "JobRunningMatchedRatio": {
        "description": "running / ( running + matched + received + checking ) jobs ( 30 min )",
        "module": "JobRunningMatchedRatioPolicy",
        "command": ("JobCommand", "JobCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    "JobRunningWaitingRatio": {
        "description": "running / ( running + waiting + staging ) jobs ( 30 min )",
        "module": "JobRunningWaitingRatioPolicy",
        "command": ("JobCommand", "JobCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    # Pilot POLICIES..............................................................
    "PilotInstantEfficiency": {
        "description": "Pilots Instant Efficiency ( 30 min )",
        "module": "PilotEfficiencyPolicy",
        "command": ("PilotCommand", "PilotCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    # Site status propagation POLICIES..............................................................
    "PropagationPolicy": {
        "description": "Site status propagation",
        "module": "PropagationPolicy",
        "command": ("PropagationCommand", "PropagationCommand"),
        "args": {"onlyCache": True, "timespan": 1800},
    },
    # ALWAYS SOMETHING POLICIES...................................................
    "AlwaysActive": {
        "description": "A Policy that always returns Active",
        "module": "AlwaysActivePolicy",
        "command": None,
        "args": None,
    },
    "AlwaysDegraded": {
        "description": "A Policy that always returns Degraded",
        "module": "AlwaysDegradedPolicy",
        "command": None,
        "args": None,
    },
    "AlwaysProbing": {
        "description": "A Policy that always returns Probing",
        "module": "AlwaysProbingPolicy",
        "command": None,
        "args": None,
    },
    "AlwaysBanned": {
        "description": "A Policy that always returns Banned",
        "module": "AlwaysBannedPolicy",
        "command": None,
        "args": None,
    },
}



""" The PBS TimeLeft utility interrogates the PBS batch system for the
    current CPU and Wallclock consumed, as well as their limits.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import os
import re
import time

from DIRAC import S_OK, S_ERROR
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.TimeLeft import runCommand
from DIRAC.Resources.Computing.BatchSystems.TimeLeft.ResourceUsage import ResourceUsage


class PBSResourceUsage(ResourceUsage):
    """
    This is the PBS plugin of the TimeLeft Utility
    """

    def __init__(self):
        """Standard constructor"""
        super(PBSResourceUsage, self).__init__("PBS", "PBS_JOBID")

        self.queue = os.environ.get("PBS_O_QUEUE")
        pbsPath = os.environ.get("PBS_O_PATH")
        if pbsPath:
            os.environ["PATH"] += ":" + pbsPath

        self.log.verbose("PBS_JOBID=%s, PBS_O_QUEUE=%s" % (self.jobID, self.queue))
        self.startTime = time.time()

    #############################################################################
    def getResourceUsage(self):
        """Returns S_OK with a dictionary containing the entries CPU, CPULimit,
        WallClock, WallClockLimit, and Unit for current slot.
        """
        cmd = "qstat -f %s" % (self.jobID)
        result = runCommand(cmd)
        if not result["OK"]:
            return result

        cpu = None
        cpuLimit = None
        wallClock = None
        wallClockLimit = None

        lines = str(result["Value"]).split("\n")
        for line in lines:
            info = line.split()
            if re.search(".*resources_used.cput.*", line):
                if len(info) >= 3:
                    cpuList = info[2].split(":")
                    newcpu = (float(cpuList[0]) * 60 + float(cpuList[1])) * 60 + float(cpuList[2])
                    if not cpu or newcpu > cpu:
                        cpu = newcpu
                else:
                    self.log.warn('Problem parsing "%s" for CPU consumed' % line)
            if re.search(".*resources_used.pcput.*", line):
                if len(info) >= 3:
                    cpuList = info[2].split(":")
                    newcpu = (float(cpuList[0]) * 60 + float(cpuList[1])) * 60 + float(cpuList[2])
                    if not cpu or newcpu > cpu:
                        cpu = newcpu
                else:
                    self.log.warn('Problem parsing "%s" for CPU consumed' % line)
            if re.search(".*resources_used.walltime.*", line):
                if len(info) >= 3:
                    wcList = info[2].split(":")
                    wallClock = (float(wcList[0]) * 60 + float(wcList[1])) * 60 + float(wcList[2])
                else:
                    self.log.warn('Problem parsing "%s" for elapsed wall clock time' % line)
            if re.search(".*Resource_List.cput.*", line):
                if len(info) >= 3:
                    cpuList = info[2].split(":")
                    newcpuLimit = (float(cpuList[0]) * 60 + float(cpuList[1])) * 60 + float(cpuList[2])
                    if not cpuLimit or newcpuLimit < cpuLimit:
                        cpuLimit = newcpuLimit
                else:
                    self.log.warn('Problem parsing "%s" for CPU limit' % line)
            if re.search(".*Resource_List.pcput.*", line):
                if len(info) >= 3:
                    cpuList = info[2].split(":")
                    newcpuLimit = (float(cpuList[0]) * 60 + float(cpuList[1])) * 60 + float(cpuList[2])
                    if not cpuLimit or newcpuLimit < cpuLimit:
                        cpuLimit = newcpuLimit
                else:
                    self.log.warn('Problem parsing "%s" for CPU limit' % line)
            if re.search(".*Resource_List.walltime.*", line):
                if len(info) >= 3:
                    wcList = info[2].split(":")
                    wallClockLimit = (float(wcList[0]) * 60 + float(wcList[1])) * 60 + float(wcList[2])
                else:
                    self.log.warn('Problem parsing "%s" for wall clock limit' % line)

        consumed = {"CPU": cpu, "CPULimit": cpuLimit, "WallClock": wallClock, "WallClockLimit": wallClockLimit}
        self.log.debug(consumed)

        if None not in consumed.values():
            self.log.debug("TimeLeft counters complete:", str(consumed))
            return S_OK(consumed)
        missed = [key for key, val in consumed.items() if val is None]
        self.log.info("Could not determine parameter", ",".join(missed))
        self.log.info("This is the stdout from the batch system call\n%s" % (result["Value"]))

        if cpuLimit or wallClockLimit:
            # We have got a partial result from PBS, assume that we ran for too short time
            if not cpuLimit:
                consumed["CPULimit"] = wallClockLimit * 0.8
            if not wallClockLimit:
                consumed["WallClockLimit"] = cpuLimit / 0.8
            if not cpu:
                consumed["CPU"] = int(time.time() - self.startTime)
            if not wallClock:
                consumed["WallClock"] = int(time.time() - self.startTime)
            self.log.verbose("TimeLeft counters restored:", str(consumed))
            return S_OK(consumed)
        msg = "Could not determine some parameters"
        self.log.info(msg, ":\nThis is the stdout from the batch system call\n%s" % (result["Value"]))
        retVal = S_ERROR(msg)
        retVal["Value"] = consumed
        return retVal



""" Transition methods to allow to move from DEncode to JEncode

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from DIRAC.Core.Utilities import DEncode, JEncode

__RCSID__ = "$Id$"


def encode(inData):
    """Encode the input data

    :param inData: data to be encoded

    :return: an encoded string
    """
    if os.getenv("DIRAC_USE_JSON_ENCODE", "NO").lower() in ("yes", "true"):
        return JEncode.encode(inData)
    return DEncode.encode(inData)


def decode(encodedData):
    """Decode the encoded string

    :param encodedData: encoded string

    :return: the decoded objects, encoded object length

    """
    if os.getenv("DIRAC_USE_JSON_DECODE", "Yes").lower() in ("yes", "true"):
        try:
            # 'null' is a special case.
            # None is encoded as 'null' as JSON
            # that DEncode would understand as None, since it starts with 'n'
            # but the length of the decoded string will not be correct
            if encodedData == "null":
                raise Exception
            return DEncode.decode(encodedData)
        except Exception:
            return JEncode.decode(encodedData)
    return DEncode.decode(encodedData)



#!/usr/bin/env python
"""
System administrator client.

Example:
  $ dirac-admin-sysadmin-cli --host dirac.in2p3.fr
  DIRAC Root Path = /afs/in2p3.fr/home/h/hamar/DIRAC-v5r12
  dirac.in2p3.fr >
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script


@Script()
def main():
    host = None
    Script.registerSwitch("H:", "host=", "   Target host")
    Script.parseCommandLine(ignoreErrors=False)
    for switch in Script.getUnprocessedSwitches():
        if switch[0].lower() == "h" or switch[0].lower() == "host":
            host = switch[1]

    from DIRAC.FrameworkSystem.Client.SystemAdministratorClientCLI import SystemAdministratorClientCLI

    cli = SystemAdministratorClientCLI(host)
    cli.cmdloop()


if __name__ == "__main__":
    main()



""" BaseHistoryCorrector is a base class for correctors of user shares within
    a given group based on the history of the resources consumption
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time as nativetime

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getDNForUsername
from DIRAC.WorkloadManagementSystem.private.correctors.BaseCorrector import BaseCorrector


class BaseHistoryCorrector(BaseCorrector):

    _GLOBAL_MAX_CORRECTION = "MaxGlobalCorrection"
    _SLICE_TIME_SPAN = "TimeSpan"
    _SLICE_WEIGHT = "Weight"
    _SLICE_MAX_CORRECTION = "MaxCorrection"

    def initialize(self):
        self.log = gLogger.getSubLogger("HistoryCorrector")
        self.__usageHistory = {}
        self.__slices = {}
        self.__lastHistoryUpdate = 0
        self.__globalCorrectionFactor = 5
        self._fillSlices()
        return S_OK()

    def _fillSlices(self):
        self.log.info("Filling time slices...")
        self.__slices = {}
        self.__globalCorrectionFactor = self.getCSOption(self._GLOBAL_MAX_CORRECTION, 5)
        result = self.getCSSections()
        if not result["OK"]:
            self.log.error("Cound not get configured time slices", result["Message"])
            return
        timeSlices = result["Value"]
        for timeSlice in timeSlices:
            self.__slices[timeSlice] = {}
            for key, defaultValue in (
                (self._SLICE_TIME_SPAN, 604800),
                (self._SLICE_WEIGHT, 1),
                (self._SLICE_MAX_CORRECTION, 3),
            ):
                self.__slices[timeSlice][key] = self.getCSOption("%s/%s" % (timeSlice, key), defaultValue)
        # Weight has to be normalized to sum 1
        weightSum = 0
        for timeSlice in self.__slices:
            weightSum += self.__slices[timeSlice][self._SLICE_WEIGHT]
        for timeSlice in self.__slices:
            self.__slices[timeSlice][self._SLICE_WEIGHT] /= float(weightSum)
        self.log.info("Found %s time slices" % len(self.__slices))

    def updateHistoryKnowledge(self):
        updatePeriod = self.getCSOption("UpdateHistoryPeriod", 900)
        now = nativetime.time()
        if self.__lastHistoryUpdate + updatePeriod > now:
            self.log.verbose("Skipping history update. Last update was less than %s secs ago" % updatePeriod)
            return
        self.__lastHistoryUpdate = now
        self.log.info("Updating history knowledge")
        self.__usageHistory = {}
        for timeSlice in self.__slices:
            result = self._getUsageHistoryForTimeSpan(self.__slices[timeSlice][self._SLICE_TIME_SPAN], self.getGroup())
            if not result["OK"]:
                self.__usageHistory = {}
                self.log.warn("Could not get history for slice", "%s: %s" % (timeSlice, result["Message"]))
                return
            self.__usageHistory[timeSlice] = result["Value"]
            self.log.verbose(
                "Got history for slice %s (%s entities in slice)" % (timeSlice, len(self.__usageHistory[timeSlice]))
            )
        self.log.info("Updated history knowledge")

    def _getHistoryData(self, _timeSpan, _groupToUse):
        """Get history data from an external source to be defined in a derived class

        :param int timeSpan: time span
        :param str groupToUse: requested user group
        :return: dictionary with history data
        """
        return S_ERROR("Not implemented !")

    def _getUsageHistoryForTimeSpan(self, timeSpan, groupToUse=""):

        result = self._getHistoryData(timeSpan, groupToUse)

        if not result["OK"]:
            self.log.error("Cannot get history data", result["Message"])
            return result
        data = result["Value"].get("data", [])
        if not data:
            message = "Empty history data"
            self.log.warn(message)
            return S_ERROR(message)

        # Map the usernames to DNs
        if groupToUse:
            mappedData = {}
            for userName in data:
                result = getDNForUsername(userName)
                if not result["OK"]:
                    self.log.error("User does not have any DN assigned", "%s :%s" % (userName, result["Message"]))
                    continue
                for userDN in result["Value"]:
                    mappedData[userDN] = data[userName]
            data = mappedData

        return S_OK(data)

    def __normalizeShares(self, entityShares):
        totalShare = 0.0
        normalizedShares = {}
        # Normalize shares
        for entity in entityShares:
            totalShare += entityShares[entity]
        self.log.verbose("Total share for given entities is %.3f" % totalShare)
        for entity in entityShares:
            normalizedShare = entityShares[entity] / totalShare
            normalizedShares[entity] = normalizedShare
            self.log.verbose("Normalized share for %s: %.3f" % (entity, normalizedShare))

        return normalizedShares

    def applyCorrection(self, entitiesExpectedShare):
        # Normalize expected shares
        normalizedShares = self.__normalizeShares(entitiesExpectedShare)

        if not self.__usageHistory:
            self.log.verbose("No history knowledge available. Correction is 1 for all entities")
            return entitiesExpectedShare

        entitiesSliceCorrections = dict([(entity, []) for entity in entitiesExpectedShare])
        for timeSlice in self.__usageHistory:
            self.log.verbose("Calculating correction for time slice %s" % timeSlice)
            sliceTotal = 0.0
            sliceHistory = self.__usageHistory[timeSlice]
            for entity in entitiesExpectedShare:
                if entity in sliceHistory:
                    sliceTotal += sliceHistory[entity]
                    self.log.verbose("Usage for %s: %.3f" % (entity, sliceHistory[entity]))
            self.log.verbose("Total usage for slice %.3f" % sliceTotal)
            if sliceTotal == 0.0:
                self.log.verbose("Slice usage is 0, skeeping slice")
                continue
            maxSliceCorrection = self.__slices[timeSlice][self._SLICE_MAX_CORRECTION]
            minSliceCorrection = 1.0 / maxSliceCorrection
            for entity in entitiesExpectedShare:
                if entity in sliceHistory:
                    normalizedSliceUsage = sliceHistory[entity] / sliceTotal
                    self.log.verbose(
                        "Entity %s is present in slice %s (normalized usage %.2f)"
                        % (entity, timeSlice, normalizedSliceUsage)
                    )
                    sliceCorrectionFactor = normalizedShares[entity] / normalizedSliceUsage
                    sliceCorrectionFactor = min(sliceCorrectionFactor, maxSliceCorrection)
                    sliceCorrectionFactor = max(sliceCorrectionFactor, minSliceCorrection)
                    sliceCorrectionFactor *= self.__slices[timeSlice][self._SLICE_WEIGHT]
                else:
                    self.log.verbose("Entity %s is not present in slice %s" % (entity, timeSlice))
                    sliceCorrectionFactor = maxSliceCorrection
                self.log.verbose("Slice correction factor for entity %s is %.3f" % (entity, sliceCorrectionFactor))
                entitiesSliceCorrections[entity].append(sliceCorrectionFactor)

        correctedEntityShare = {}
        maxGlobalCorrectionFactor = self.__globalCorrectionFactor
        minGlobalCorrectionFactor = 1.0 / maxGlobalCorrectionFactor
        for entity in entitiesSliceCorrections:
            entityCorrectionFactor = 0.0
            slicesCorrections = entitiesSliceCorrections[entity]
            if not slicesCorrections:
                self.log.verbose("Entity does not have any correction %s" % entity)
                correctedEntityShare[entity] = entitiesExpectedShare[entity]
            else:
                for cF in entitiesSliceCorrections[entity]:
                    entityCorrectionFactor += cF
                entityCorrectionFactor = min(entityCorrectionFactor, maxGlobalCorrectionFactor)
                entityCorrectionFactor = max(entityCorrectionFactor, minGlobalCorrectionFactor)
                correctedShare = entitiesExpectedShare[entity] * entityCorrectionFactor
                correctedEntityShare[entity] = correctedShare
                self.log.verbose(
                    "Final correction factor for entity %s is %.3f\n Final share is %.3f"
                    % (entity, entityCorrectionFactor, correctedShare)
                )
        self.log.verbose(
            "Initial shares:\n  %s"
            % "\n  ".join(["%s : %.2f" % (en, entitiesExpectedShare[en]) for en in entitiesExpectedShare])
        )
        self.log.verbose(
            "Corrected shares:\n  %s"
            % "\n  ".join(["%s : %.2f" % (en, correctedEntityShare[en]) for en in correctedEntityShare])
        )
        return correctedEntityShare
